""".
https://googleapis.github.io/google-api-python-client/docs/dyn/analyticsreporting_v4.html
"""
import logging
from typing import Any, Dict, List, Literal, Optional

from pydantic import StrictStr, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel
from nr_ops.ops.ops.connector_ops.interfaces.base import validate_hook_type_and_config

logger = logging.getLogger(__name__)


class GoogleAnalyticsConnOpConfigModel(BaseOpConfigModel):
    hook_type: Literal[
        "connector.hooks.airflow_gcp_hook",
        "connector.hooks.gcp_service_account_from_file",
    ]
    hook_config: Dict[StrictStr, Any]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    _validate_hook_type_and_config = root_validator(allow_reuse=True)(
        validate_hook_type_and_config
    )


class GoogleAnalyticsConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAnalyticsConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAnalyticsConnOp(BaseConnectorOp):
    OP_TYPE = "connector.google_analytics"
    OP_CONFIG_MODEL = GoogleAnalyticsConnOpConfigModel
    OP_METADATA_MODEL = GoogleAnalyticsConnOpMetadataModel
    OP_AUDIT_MODEL = GoogleAnalyticsConnOpAuditModel

    templated_fields = None

    def __init__(self, hook_type: str, hook_config: Dict[str, Any], **kwargs):
        self.hook_type = hook_type
        self.hook_config = hook_config
        self.templated_fields = kwargs.get("templated_fields", [])

        from nr_ops.ops.op_collection import OP_COLLECTION

        self.hook_op = OP_COLLECTION[self.hook_type](**self.hook_config)
        self.hook = self.hook_op.run().data

        # Will be populated in run()
        self.analytics: Any = None

    def get_report(
        self,
        view_id: str,
        date_ranges: Optional[List[Dict[str, str]]] = None,
        metrics: Optional[List[Dict[str, str]]] = None,
        dimensions: Optional[List[Dict[str, str]]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """."""
        if (
            self.hook_type == "connector.hooks.airflow_gcp_hook"
            or self.hook_type == "connector.hooks.gcp_service_account_from_file"
        ):
            payload = {"reportRequests": [{"viewId": view_id}]}
            if date_ranges:
                payload["reportRequests"][0]["dateRanges"] = date_ranges
            if metrics:
                payload["reportRequests"][0]["metrics"] = metrics
            if dimensions:
                payload["reportRequests"][0]["dimensions"] = dimensions

            results = self.analytics.reports().batchGet(body=payload).execute()

            if len(results["reports"]) > 1:
                raise NotImplementedError("Expected only 1 report.")

            results = results["reports"][0]
            data = [
                {
                    "metrics": {
                        metric["name"]: row["metrics"][i]["values"][0]
                        for i, metric in enumerate(
                            results["columnHeader"]["metricHeader"][
                                "metricHeaderEntries"
                            ]
                        )
                    },
                    "dimensions": {
                        dimension_name: row["dimensions"][i]
                        for i, dimension_name in enumerate(
                            results["columnHeader"]["dimensions"]
                        )
                    },
                }
                for row in results["data"]["rows"]
            ]
            output = {
                "ga_type": "universal",
                "request": {
                    "view_id": view_id,
                    "date_ranges": date_ranges,
                    "metrics": metrics,
                    "dimensions": dimensions,
                },
                "response": {
                    "view_id": view_id,
                    "date_ranges": date_ranges,
                    "data": data,
                },
            }
            return output
        else:
            raise NotImplementedError()

    def run(self) -> OpMsg:
        """."""
        logger.info(f"GoogleAnalyticsConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="GoogleAnalyticsConnOp.run:"
        )

        # Source: https://googleapis.github.io/google-api-python-client/docs/dyn/analyticsreporting_v4.html
        if self.hook_type == "connector.hooks.airflow_gcp_hook":
            # Type annotate for pycharm/mypy/autocomplete
            from airflow.providers.google.common.hooks.base_google import GoogleBaseHook

            self.hook: GoogleBaseHook

            import google.auth.credentials
            from googleapiclient.discovery import build

            credentials: google.auth.credentials.Credentials = (
                self.hook._get_credentials()
            )
            self.analytics = build("analyticsreporting", "v4", credentials=credentials)
        elif self.hook_type == "connector.hooks.gcp_service_account_from_file":
            # Type annotate for pycharm/mypy/autocomplete
            from googleapiclient.discovery import build
            from oauth2client.service_account import ServiceAccountCredentials

            credentials: ServiceAccountCredentials = self.hook.credentials
            self.analytics = build("analyticsreporting", "v4", credentials=credentials)

        else:
            raise NotImplementedError()

        return OpMsg(
            data=self,
            metadata=GoogleAnalyticsConnOpMetadataModel(),
            audit=GoogleAnalyticsConnOpAuditModel(),
        )



import logging
from typing import Any, Dict, Generator, List, Optional

from googleapiclient.discovery import build
from pydantic import StrictBool, StrictInt, StrictStr, conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.google_sheets import GoogleSheetsConnOp

logger = logging.getLogger(__name__)


class GoogleSheetsGetCellRangeOpConfigModel(BaseOpConfigModel):
    google_sheets_conn_id: StrictStr
    spreadsheet_id: StrictStr
    sheet_name: StrictStr
    cell_range: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleSheetsGetCellRangeOpMetadataModel(BaseOpMetadataModel):
    spreadsheet_id: StrictStr
    sheet_name: StrictStr
    cell_range: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleSheetsGetCellRangeOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleSheetsGetCellRangeOp(BaseGeneratorOp):
    OP_TYPE = "generator.google_sheets.get_cell_range"
    OP_CONFIG_MODEL = GoogleSheetsGetCellRangeOpConfigModel
    OP_METADATA_MODEL = GoogleSheetsGetCellRangeOpMetadataModel
    OP_AUDIT_MODEL = GoogleSheetsGetCellRangeOpAuditModel

    templated_fields = None

    def __init__(
        self,
        google_sheets_conn_id: str,
        spreadsheet_id: str,
        sheet_name: str,
        cell_range: str,
        **kwargs,
    ):
        """."""
        self.google_sheets_conn_id = google_sheets_conn_id
        self.spreadsheet_id = spreadsheet_id
        self.sheet_name = sheet_name
        self.cell_range = cell_range
        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.conn: GoogleSheetsConnOp = op_manager.get_connector(
            op_id=self.google_sheets_conn_id
        )
        self.sheet_service = self.conn.sheet_service

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"GoogleSheetsGetCellRangeOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="GoogleSheetsGetCellRangeOp.run:"
        )

        range_ = f"{self.sheet_name}!{self.cell_range}"

        logger.info(
            f"GoogleSheetsGetCellRangeOp.run: Fetching cell range from "
            f"{self.spreadsheet_id=} and merged sheet & range {range_=}."
        )

        result = (
            self.sheet_service.values()
            .get(spreadsheetId=self.spreadsheet_id, range=range_)
            .execute()
        )

        rows: List[List] = result.get("values", [])

        logger.info(
            f"GoogleSheetsGetCellRangeOp.run: Done fetching cell range from "
            f"{self.spreadsheet_id=} and {range_=} (merged sheet & range). "
            f"n_rows={len(rows)}."
        )

        if not rows:
            logger.info(f"GoogleSheetsGetCellRangeOp.run: No data. {rows=}. Consuming.")
            return

        logger.info(
            f"GoogleSheetsGetCellRangeOp.run: Data exists. "
            f"n_rows={len(rows)}. Yielding."
        )

        yield OpMsg(
            data=rows,
            metadata=GoogleSheetsGetCellRangeOpMetadataModel(
                spreadsheet_id=self.spreadsheet_id,
                sheet_name=self.sheet_name,
                cell_range=self.cell_range,
            ),
            audit=GoogleSheetsGetCellRangeOpAuditModel(),
        )



import logging
import os
from typing import Any, Dict, List, Optional

from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp
from nr_ops.ops.op import Op

logger = logging.getLogger(__name__)


class OpManager(object):
    def __init__(self):
        """."""
        # Currently it seems appropriate to store the environment variables in the
        # op_manager and not withing OpSubManager.
        self._env_vars = {}
        self.__ops = {}
        self.__metadata = {}
        self.__data = {}

    def set_env_vars(self, env_vars: Optional[Dict[str, str]] = None):
        """Store environment variables from os.env."""
        env_vars = {} if env_vars is None else env_vars

        if not env_vars:
            return

        env_vars_keys = list(env_vars.keys())
        logger.info(f"OpManager: Settings environment variables. {env_vars_keys=}")

        for env_var_key, env_var_value in env_vars.items():

            if not isinstance(env_var_value, str):
                raise TypeError(
                    f"set_env_vars: The environment variable value for the "
                    f"{env_var_key=} is not a string. Please check the config file "
                    f"and make sure that ALL environment variable values are strings."
                )

            # Set the environment variable both in os.environ and self.env_vars
            os.environ[env_var_key] = env_var_value
            self.env_vars[env_var_key] = env_var_value

    def import_env_vars(self, env_vars: Optional[List[str]] = None):
        """Store environment variables from os.env."""
        logger.info(f"OpManager: Importing environment variables. {env_vars=}")
        if env_vars is None:
            return

        for env_var in env_vars:
            if env_var not in os.environ:
                raise ValueError(
                    f"{env_var=} not found in os.environ! Please check for typos in "
                    f"`import_env_vars` in the config file or make sure the "
                    f"environment variable is set."
                )
            self.env_vars[env_var] = os.getenv(env_var)

    @property
    def env_vars(self):
        return self._env_vars

    @property
    def metadata(self):
        return self.__metadata

    @property
    def data(self):
        return self.__data

    @property
    def ops(self):
        return self.__ops

    def store_metadata(self, op: Op, msg: OpMsg):
        if op.op_id is None:
            # NOT registering op with op_id=None
            return
        self.__metadata[op.op_id] = msg.metadata

    def store_data(self, op: Op, msg: OpMsg):
        if op.op_id is None:
            # NOT registering op with op_id=None
            return
        self.__data[op.op_id] = msg.data

    def store_op(self, op: Op):
        if op.op_id is None:
            # NOT registering op with op_id=None
            return
        self.__ops[op.op_id] = op

    def get_data(self, op_id: str) -> Any:
        return self.__data[op_id]

    def get_metadata(self, op_id: str) -> Any:
        return self.__metadata[op_id]

    # Alias
    def get_connector(self, op_id: str) -> BaseConnectorOp:
        conn_op = self.__data[op_id]
        if not isinstance(conn_op, BaseConnectorOp):
            raise ValueError(
                f"Expected {op_id=} to be of type BaseConnectorOp instead "
                f"received {type(conn_op)=}."
            )
        return self.__data[op_id]


GLOBAL_OP_MANAGERS: Optional[OpManager] = None


def init_global_op_manager() -> OpManager:
    """."""
    global GLOBAL_OP_MANAGERS
    GLOBAL_OP_MANAGERS = OpManager()
    return GLOBAL_OP_MANAGERS


def get_global_op_manager() -> OpManager:
    """."""
    if GLOBAL_OP_MANAGERS is None:
        raise ValueError(
            f"GLOBAL_OP_MANAGERS has not been initialized. "
            f"Please call init_global_op_managers() first."
        )

    return GLOBAL_OP_MANAGERS



import logging
import time
from typing import Generator, Optional

from pydantic import StrictBool, conint

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class SkipEmptyInputGeneratorOpConfigModel(BaseOpConfigModel):
    skip_zero_length: StrictBool = True
    skip_none: StrictBool = True

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SkipEmptyInputGeneratorOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SkipEmptyInputGeneratorOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SkipEmptyInputGeneratorOp(BaseGeneratorOp):
    OP_TYPE = "generator.skip_empty_input"
    OP_CONFIG_MODEL = SkipEmptyInputGeneratorOpConfigModel
    OP_METADATA_MODEL = SkipEmptyInputGeneratorOpMetadataModel
    OP_AUDIT_MODEL = SkipEmptyInputGeneratorOpAuditModel

    templated_fields = None

    def __init__(self, skip_zero_length: bool = True, skip_none: bool = True, **kwargs):
        self.skip_zero_length = skip_zero_length
        self.skip_none = skip_none

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"SkipEmptyInputGeneratorOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="SkipEmptyInputGeneratorOp.run:"
        )

        if self.skip_none and msg.data is None:
            logger.info(
                f"SkipEmptyInputGeneratorOp.run: Skipping input OpMsg because "
                f"{self.skip_none=} and {type(msg.data)=}"
            )
            return

        if self.skip_zero_length and len(msg.data) == 0:
            logger.info(
                f"SkipEmptyInputGeneratorOp.run: Skipping input OpMsg because "
                f"{self.skip_zero_length=} and len({msg.data})=0 for {type(msg.data)=}"
            )
            return

        logger.info(
            f"SkipEmptyInputGeneratorOp.run: Done checking input for {self.skip_zero_length=}"
            f"and {self.skip_none=}. msg.data is not empty. Yielding OpMsg."
        )

        yield OpMsg(
            data=msg.data,
            metadata=SkipEmptyInputGeneratorOpMetadataModel(),
            audit=SkipEmptyInputGeneratorOpAuditModel(),
        )



import logging

logger = logging.getLogger(__name__)


def custom_raise_exception(msg: str):
    """A function that accepts as string msg and raises an Exception with that msg.

    This is a useful function to use in the eval() function since eval() does not
    allow raising exceptions using the `raise` keyword.

    Args:
        msg (str): The message to raise.
    """
    raise Exception(msg)


def custom_generate_batches_from_list(data: list, batch_size: int):
    """A function that accepts a list and batch size and returns a generator that
    yields batches from the input list.

    Args:
        data (list): The list to batch.
        batch_size (int): The size of each batch.

    Yields:
        list: A batch of items from input list.
    """
    logger.info(
        f"custom_generate_batches_from_list: Batching iterable of {type(data)=} "
        f"and {len(data)=} into batches of {batch_size=}."
    )
    for i in range(0, len(data), batch_size):
        yield data[i : i + batch_size]



"""."""
import logging
from typing import Any, Dict, Generator, List, Literal, Optional

from pydantic import BaseModel, StrictStr, conlist, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel
from nr_ops.ops.ops.connector_ops.interfaces.base import validate_hook_type_and_config

logger = logging.getLogger(__name__)


class MysqlConnOpConfigModel(BaseOpConfigModel):
    hook_type: Literal[
        "connector.hooks.airflow_mysql_hook",
    ]
    hook_config: Dict[StrictStr, Any]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    _validate_hook_type_and_config = root_validator(allow_reuse=True)(
        validate_hook_type_and_config
    )


class MysqlConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MysqlConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MysqlConnOp(BaseConnectorOp):
    OP_TYPE = "connector.mysql"
    OP_CONFIG_MODEL = MysqlConnOpConfigModel
    OP_METADATA_MODEL = MysqlConnOpMetadataModel
    OP_AUDIT_MODEL = MysqlConnOpAuditModel

    templated_fields = None

    def __init__(self, hook_type: str, hook_config: Dict[str, Any], **kwargs):
        self.hook_type = hook_type
        self.hook_config = hook_config
        self.templated_fields = kwargs.get("templated_fields", [])

        from nr_ops.ops.op_collection import OP_COLLECTION

        self.hook_op = OP_COLLECTION[self.hook_type](**self.hook_config)
        self.hook = self.hook_op.run().data
        self.engine = None

        if self.hook_type == "connector.hooks.airflow_mysql_hook":
            from airflow.providers.mysql.hooks.mysql import MySqlHook

            self.hook: MySqlHook = self.hook
        else:
            raise NotImplementedError()

    def get_engine(self):
        """."""
        if self.hook_type == "connector.hooks.airflow_mysql_hook":
            if self.engine is None:
                self.engine = self.hook.get_sqlalchemy_engine()

            return self.engine
        else:
            raise NotImplementedError()

    def get_connection(self):
        """."""
        if self.hook_type == "connector.hooks.airflow_mysql_hook":
            return self.hook.get_conn()
        else:
            raise NotImplementedError()

    def run(self) -> OpMsg:
        """."""
        logger.info(f"MysqlConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(time_step=None, msg=None, log_prefix="MysqlConnOp.run:")

        if self.hook_type == "connector.hooks.airflow_mysql_hook":
            pass
        else:
            raise NotImplementedError()

        return OpMsg(
            data=self,
            metadata=MysqlConnOpMetadataModel(),
            audit=MysqlConnOpAuditModel(),
        )



from __future__ import annotations

import logging
from typing import Any, Dict, Literal

from pydantic import StrictStr, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel
from nr_ops.ops.ops.connector_ops.hooks.python_counter import PythonCounterHookConnOp
from nr_ops.ops.ops.connector_ops.interfaces.base import validate_hook_type_and_config

logger = logging.getLogger(__name__)


class CounterConnOpConfigModel(BaseOpConfigModel):
    hook_type: Literal[
        "connector.hooks.python.counter",
    ]
    hook_config: Dict[StrictStr, Any]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    _validate_hook_type_and_config = root_validator(allow_reuse=True)(
        validate_hook_type_and_config
    )


class CounterConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class CounterConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class CounterConnOp(BaseConnectorOp):
    OP_TYPE = "connector.counter"
    OP_CONFIG_MODEL = CounterConnOpConfigModel
    OP_METADATA_MODEL = CounterConnOpMetadataModel
    OP_AUDIT_MODEL = CounterConnOpAuditModel

    templated_fields = None

    def __init__(self, hook_type: str, hook_config: Dict[str, Any], **kwargs):
        self.hook_type = hook_type
        self.hook_config = hook_config
        self.templated_fields = kwargs.get("templated_fields", [])

        from nr_ops.ops.op_collection import OP_COLLECTION

        self.hook_op = OP_COLLECTION[self.hook_type](**self.hook_config)
        self.hook = self.hook_op.run().data

        if self.hook_type == "connector.hooks.python.counter":
            self.hook: PythonCounterHookConnOp
        else:
            raise NotImplementedError()

    # NOTE: Returns Union[PythonCounterHookConnOp] but the interface is CounterConnOp
    def get_reference(self) -> CounterConnOp:
        """."""
        if self.hook_type == "connector.hooks.python.counter":
            return self.hook.get_reference()
        else:
            raise NotImplementedError()

    # NOTE: Returns Union[PythonCounterHookConnOp] but the interface is CounterConnOp
    def get_reference_and_reinit(self) -> CounterConnOp:
        """."""
        if self.hook_type == "connector.hooks.python.counter":
            return self.hook.get_reference_and_reinit()
        else:
            raise NotImplementedError()

    def put_update_dict(self, item: Any):
        """."""
        if self.hook_type == "connector.hooks.python.counter":
            self.hook.put_update_dict(item=item)
        else:
            raise NotImplementedError()

    def put_update_list_of_dicts(self, item: Any):
        """."""
        if self.hook_type == "connector.hooks.python.counter":
            self.hook.put_update_list_of_dicts(item=item)
        else:
            raise NotImplementedError()

    def run(self) -> OpMsg:
        """."""
        logger.info(f"CounterConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(time_step=None, msg=None, log_prefix="CounterConnOp.run:")

        if self.hook_type == "connector.hooks.python.counter":
            pass
        else:
            raise NotImplementedError()

        return OpMsg(
            data=self,
            metadata=CounterConnOpMetadataModel(),
            audit=CounterConnOpAuditModel(),
        )



import logging
from typing import Generator, List, Optional

from pydantic import conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGroupOp, BaseOp, BaseOpConfigModel
from nr_ops.ops.op import Op, OpModel

logger = logging.getLogger(__name__)


class OpChainBranchGroupOpConfigModel(BaseOpConfigModel):
    ops: conlist(OpModel, min_items=1)

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    def get_self_and_child_op_ids(self) -> Generator[str, None, None]:
        """."""
        for op in self.ops:
            yield from op.get_self_and_child_op_ids()


class OpChainBranchGroupOpMetadataModel(BaseOpMetadataModel):
    output_metadata: BaseOpMetadataModel

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OpChainBranchGroupOpAuditModel(BaseOpAuditModel):
    output_audit: BaseOpAuditModel

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OpChainBranchGroupOp(BaseGroupOp):
    """.

    Same as op_chain except that it yield the "input" msg instead of yielding "output"
    msg of the final op in the chain.
    """

    OP_TYPE = "group.op_chain_branch"
    OP_CONFIG_MODEL = OpChainBranchGroupOpConfigModel
    OP_METADATA_MODEL = OpChainBranchGroupOpMetadataModel
    OP_AUDIT_MODEL = OpChainBranchGroupOpAuditModel

    templated_fields = None

    def __init__(self, ops: List[BaseOp], **kwargs):
        self.ops = [Op(**op) for op in ops]

    @staticmethod
    def get_output_msgs_from_op(
        time_step: TimeStep,
        msgs: Generator[Optional[OpMsg], None, None],
        op: Op,
    ) -> Generator[Optional[OpMsg], None, None]:
        """A generator of output messages for each input messages.

        - All output messages of all input messages are yielded in order in one
          single generator. If the final op is a generator, then just consume the
          messages.
        """
        for input_msg in msgs:
            for output_msg in op.run(time_step=time_step, msg=input_msg):
                yield output_msg

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[Optional[OpMsg], None, None]:
        """."""
        logger.info(f"OpChainBranchGroupOp.run: Running")
        # Convert a single message into a generator of messages.
        # This is just to treat all messages as if they came from a generator.
        msgs = (_msg for _msg in [msg])  # type: Generator[Optional[OpMsg], None, None]

        for op in self.ops:
            # Create a chain of generators.
            # Each output generator is the input to the next generator.
            msgs = self.get_output_msgs_from_op(time_step=time_step, msgs=msgs, op=op)

        # The final generator is the generator of this op_group.
        final_msgs = msgs

        for output_msg_i, output_msg in enumerate(final_msgs):
            # CONSUME ALL MESSAGES FROM FINAL OP
            logger.info(
                f"OpChainBranchGroupOp.run: Consuming final output_msg from op_chain | "
                f"{output_msg_i=} | {type(output_msg.data)=}"
            )
            pass

        # YIELD INPUT MESSAGE
        yield msg



import logging
import re
from typing import Generator, List, Optional

from pydantic import StrictStr, validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.gcp_gcs import GCPGCSConnOp

logger = logging.getLogger(__name__)


class GCPGCSListKeysOpConfigModel(BaseOpConfigModel):
    gcs_conn_id: StrictStr
    bucket: StrictStr
    prefix: StrictStr
    regex: Optional[StrictStr] = None
    iterate_over_keys: bool = True

    @validator("prefix", pre=False)
    def validate_prefix(cls, prefix: str):
        if not prefix.endswith("/"):
            raise ValueError(f"{prefix=} must end with a '/'")

        if prefix.startswith("/"):
            raise ValueError(f"{prefix=} must not start with a '/'")

        return prefix

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSListKeysOpMetadataModel(BaseOpMetadataModel):
    bucket: StrictStr
    prefix: StrictStr
    key: Optional[StrictStr] = None
    keys: Optional[List[StrictStr]] = None
    filename: Optional[StrictStr] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSListKeysOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSListKeysOp(BaseGeneratorOp):
    OP_TYPE = "generator.gcp.gcs.list_keys"
    OP_CONFIG_MODEL = GCPGCSListKeysOpConfigModel
    OP_METADATA_MODEL = GCPGCSListKeysOpMetadataModel
    OP_AUDIT_MODEL = GCPGCSListKeysOpAuditModel

    templated_fields = None

    def __init__(
        self,
        gcs_conn_id: str,
        bucket: str,
        prefix: str,
        regex: Optional[str] = None,
        iterate_over_keys: bool = True,
        **kwargs,
    ):
        """."""
        self.gcs_conn_id = gcs_conn_id
        self.bucket = bucket
        self.prefix = prefix
        self.regex = regex
        self.iterate_over_keys = iterate_over_keys

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.gcs_conn: GCPGCSConnOp = op_manager.get_connector(op_id=self.gcs_conn_id)

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"GCPGCSListKeysOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="GCPGCSListKeysOp.run:"
        )

        keys = self.gcs_conn.list_keys(bucket=self.bucket, prefix=self.prefix)
        logger.info(
            f"GCPGCSListKeysOp.run: {len(keys)=} found in {self.bucket=} with "
            f"{self.prefix=}. Sample keys: \n{keys[:5]}"
        )

        if self.regex is not None:
            compiled_regex = re.compile(self.regex)
            keys = [key for key in keys if compiled_regex.match(key)]
            logger.info(
                f"GCPGCSListKeysOp.run: Applied regex matching on keys. "
                f" {len(keys)=} matched with {self.regex=}. Sample keys: \n{keys[:5]}"
            )

        if self.iterate_over_keys:
            for key in keys:
                filename = key[len(self.prefix) :]
                logger.info(f"GCPGCSListKeysOp.run: Yielding {key=}, {filename=}")
                yield OpMsg(
                    data=key,
                    metadata=GCPGCSListKeysOpMetadataModel(
                        bucket=self.bucket,
                        prefix=self.prefix,
                        keys=None,
                        key=key,
                        filename=filename,
                    ),
                    audit=GCPGCSListKeysOpAuditModel(),
                )
        else:
            logger.info(
                f"GCPGCSListKeysOp.run: Yielding all keys of {len(keys)=} as output."
            )
            yield OpMsg(
                data=keys,
                metadata=GCPGCSListKeysOpMetadataModel(
                    bucket=self.bucket,
                    prefix=self.prefix,
                    keys=keys,
                    key=None,
                    filename=None,
                ),
                audit=GCPGCSListKeysOpAuditModel(),
            )



import io
import logging
import pickle
import time
from typing import Generator, Literal, Optional

from pydantic import StrictBool, conint

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class UnPickleGeneratorOpConfigModel(BaseOpConfigModel):
    input_type: Literal["bytes"]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class UnPickleGeneratorOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class UnPickleGeneratorOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class UnPickleGeneratorOp(BaseGeneratorOp):
    OP_TYPE = "generator.unpickle"
    OP_CONFIG_MODEL = UnPickleGeneratorOpConfigModel
    OP_METADATA_MODEL = UnPickleGeneratorOpMetadataModel
    OP_AUDIT_MODEL = UnPickleGeneratorOpAuditModel

    templated_fields = None

    def __init__(
        self,
        input_type: Literal["bytes", "bytesio"],
        fix_imports: Optional[StrictBool] = None,
        **kwargs,
    ):
        self.input_type = input_type
        self.fix_imports = fix_imports

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"UnPickleGeneratorOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="UnPickleGeneratorOp.run:"
        )

        logger.info(f"UnPickleGeneratorOp.run: ")

        pickle_kwargs = {}
        if self.fix_imports is not None:
            pickle_kwargs["fix_imports"] = self.fix_imports

        output = pickle.loads(msg.data, **pickle_kwargs)

        logger.info(
            f"UnPickleGeneratorOp.run: Yielding unpickled input of type "
            f"{self.input_type=}, {type(output)=}"
        )

        yield OpMsg(
            data=output,
            metadata=UnPickleGeneratorOpMetadataModel(),
            audit=UnPickleGeneratorOpAuditModel(),
        )



import logging
from typing import Any, Dict, Literal

from pydantic import StrictStr, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel
from nr_ops.ops.ops.connector_ops.interfaces.base import validate_hook_type_and_config

logger = logging.getLogger(__name__)


class S3ConnOpConfigModel(BaseOpConfigModel):
    hook_type: Literal[
        "connector.hooks.airflow_s3_hook",
    ]
    hook_config: Dict[StrictStr, Any]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    _validate_hook_type_and_config = root_validator(allow_reuse=True)(
        validate_hook_type_and_config
    )


class S3ConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class S3ConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class S3ConnOp(BaseConnectorOp):
    OP_TYPE = "connector.s3"
    OP_CONFIG_MODEL = S3ConnOpConfigModel
    OP_METADATA_MODEL = S3ConnOpMetadataModel
    OP_AUDIT_MODEL = S3ConnOpAuditModel

    templated_fields = None

    def __init__(self, hook_type: str, hook_config: Dict[str, Any], **kwargs):
        self.hook_type = hook_type
        self.hook_config = hook_config
        self.templated_fields = kwargs.get("templated_fields", [])

        from nr_ops.ops.op_collection import OP_COLLECTION

        self.hook_op = OP_COLLECTION[self.hook_type](**self.hook_config)
        self.hook = self.hook_op.run().data

        if self.hook_type == "connector.hooks.airflow_s3_hook":
            from airflow.providers.amazon.aws.hooks.s3 import S3Hook

            self.hook: S3Hook
        else:
            raise NotImplementedError()

    def list_keys(self, bucket: str, prefix: str, delimiter: str = "/"):
        """."""
        if self.hook_type == "connector.hooks.airflow_s3_hook":
            from airflow.providers.amazon.aws.hooks.s3 import S3Hook

            self.hook: S3Hook
            keys = self.hook.list_keys(
                bucket_name=bucket,
                prefix=prefix,
                delimiter=delimiter,
            )
            return keys
        else:
            raise NotImplementedError()

    def get_key(self, bucket: str, key: str, output_type: str):
        """."""
        if self.hook_type == "connector.hooks.airflow_s3_hook":
            from airflow.providers.amazon.aws.hooks.s3 import S3Hook

            self.hook: S3Hook
            transfer_obj = self.hook.get_key(bucket_name=bucket, key=key)
            if output_type == "bytes":
                return transfer_obj.get()["Body"].read()
            elif output_type == "bytesio":
                return transfer_obj.get()["Body"]
            else:
                raise NotImplementedError()
        else:
            raise NotImplementedError()

    def put_key(
        self,
        bucket: str,
        key: str,
        input_type: str,
        replace: bool,
        data: Any,
    ):
        if self.hook_type == "connector.hooks.airflow_s3_hook":
            from airflow.providers.amazon.aws.hooks.s3 import S3Hook

            self.hook: S3Hook
            if input_type == "bytes":
                self.hook.load_bytes(
                    bytes_data=data, bucket_name=bucket, key=key, replace=replace
                )
            elif input_type == "bytesio":
                self.hook.load_file_obj(
                    file_obj=data, bucket_name=bucket, key=key, replace=replace
                )
            else:
                raise NotImplementedError()
        else:
            raise NotImplementedError()

    def run(self) -> OpMsg:
        """."""
        logger.info(f"S3ConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(time_step=None, msg=None, log_prefix="S3ConnOp.run:")

        if self.hook_type == "connector.hooks.airflow_s3_hook":
            pass
        else:
            raise NotImplementedError()

        return OpMsg(
            data=self,
            metadata=S3ConnOpMetadataModel(),
            audit=S3ConnOpAuditModel(),
        )



import logging
from typing import Any, Dict, List, Union

import pandas as pd
from pydantic import BaseModel, Field, StrictBool, StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.postgres import PostgresConnOp

logger = logging.getLogger(__name__)


class IndexConfigModel(BaseModel):
    is_unique: StrictBool

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class NROpsPGCreateETLTableConsumerOpConfigModel(BaseOpConfigModel):
    postgres_conn_id: StrictStr
    schema_: StrictStr
    table: StrictStr
    # dedup_uuid is primary key
    # Force user to provide {} if they want no indexes
    indexes: Dict[StrictStr, IndexConfigModel]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False
        fields = {"schema_": "schema"}

    # Source: https://github.com/pydantic/pydantic/issues/153#issuecomment-379695371
    def dict(self, *args, **kwargs):
        d = super().dict(*args, **kwargs)
        d["schema"] = d.pop("schema_")
        return d


class NROpsPGCreateETLTableConsumerOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class NROpsPGCreateETLTableConsumerOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class NROpsPostgresCreateETLTableConsumerOp(BaseConsumerOp):
    OP_TYPE = "consumer.nr_ops.postgres_create_etl_table"
    OP_CONFIG_MODEL = NROpsPGCreateETLTableConsumerOpConfigModel
    OP_METADATA_MODEL = NROpsPGCreateETLTableConsumerOpMetadataModel
    OP_AUDIT_MODEL = NROpsPGCreateETLTableConsumerOpAuditModel

    templated_fields = None

    def __init__(
        self,
        postgres_conn_id: str,
        schema: str,
        table: str,
        indexes: Dict[str, Any],
        **kwargs,
    ):
        self.postgres_conn_id = postgres_conn_id
        self.schema = schema
        self.table = table
        self.indexes = indexes
        self.indexes_with_models = {
            key: IndexConfigModel(**value) for key, value in self.indexes.items()
        }
        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.db: Union[PostgresConnOp] = op_manager.get_connector(
            op_id=self.postgres_conn_id
        )

    def run(self, time_step: TimeStep, msg: OpMsg) -> OpMsg:
        """."""
        logger.info(f"NROpsPGCreateETLTableConsumerOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="NROpsPGCreateETLTableConsumerOp.run:",
        )

        logger.info(
            f"NROpsPGCreateETLTableConsumerOp.run: Checking to see if "
            f"'{self.schema}'.'{self.table}' already exist in the database."
        )
        # Make sure that the table does not exist!
        df = pd.read_sql(
            sql=f"""
            SELECT
                *
            FROM
                pg_catalog.pg_tables 
            WHERE 
                schemaname = '{self.schema}' 
                AND tablename = '{self.table}'
          """,
            con=self.db.get_engine(),
        )

        if len(df) == 0:
            logger.info(
                f"NROpsPGCreateETLTableConsumerOp.run: '{self.schema}'.'{self.table}' "
                f"does not exist in the database. Attempting to create the table..."
            )
        else:
            logger.info(
                f"NROpsPGCreateETLTableConsumerOp.run: '{self.schema}'.'{self.table}' "
                f"already exists in the database. Will not attempt to create table "
                f"or its indexes or triggers. Consuming..."
            )
            return OpMsg(
                data=None,
                metadata=NROpsPGCreateETLTableConsumerOpMetadataModel(),
                audit=NROpsPGCreateETLTableConsumerOpAuditModel(),
            )

        indexes = [
            f"""create """
            f"""{'unique' if index_model.is_unique else ''} """
            f"""index """
            f"""if not exists """
            f""" "{self.schema.lower().replace(' ', '_')}_{self.table.lower().replace(' ', '_')}_{index_col.lower().replace(' ', '_')}" """
            f"""on "{self.schema}"."{self.table}" ({index_col});"""
            for index_col, index_model in self.indexes_with_models.items()
        ]
        indexes_str = "\n".join(indexes)

        create_sql = f"""
            create table if not exists "{self.schema}"."{self.table}"
            (
                etl_logical_ts                       timestamp with time zone,
                etl_start_ts                         timestamp with time zone,
                etl_end_ts                           timestamp with time zone,
                etl_created_at                       timestamp with time zone default current_timestamp,
                etl_updated_at                       timestamp with time zone default current_timestamp,
                dedup_uuid                           uuid,
                data                                 JSONB,
                etl_metadata                         JSONB,
                PRIMARY KEY (dedup_uuid)
            );
    
            {indexes_str}
    
            CREATE OR REPLACE FUNCTION
                update_etl_updated_at_{self.schema.lower().replace(' ', '_')}_{self.table.lower().replace(' ', '_')}()
            RETURNS TRIGGER AS $$
            BEGIN
                NEW.etl_updated_at = now();
                RETURN NEW;
            END;
            $$ language 'plpgsql';
    
    
            DO $$ BEGIN
            DROP TRIGGER IF EXISTS
                trigger_etl_updated_at_{self.schema.lower().replace(' ', '_')}_{self.table.lower().replace(' ', '_')}
            ON
                "{self.schema}"."{self.table}";
    
            CREATE TRIGGER
                trigger_etl_updated_at_{self.schema.lower().replace(' ', '_')}_{self.table.lower().replace(' ', '_')}
            BEFORE INSERT OR UPDATE ON
                "{self.schema}"."{self.table}"
            FOR EACH ROW EXECUTE PROCEDURE
                update_etl_updated_at_{self.schema.lower().replace(' ', '_')}_{self.table.lower().replace(' ', '_')}();
            END $$;
        """

        logger.info(
            f"NROpsPGCreateETLTableConsumerOp.run: Applying the following "
            f"create statements!\n{create_sql}"
        )

        # Create table and indexes and triggers
        with self.db.get_engine().connect() as conn:
            conn.execute(create_sql)

        logger.info(
            f"NROpsPGCreateETLTableConsumerOp.run: Done creating etl table "
            f"'{self.schema}'.'{self.table}'"
        )

        return OpMsg(
            data=None,
            metadata=NROpsPGCreateETLTableConsumerOpMetadataModel(),
            audit=NROpsPGCreateETLTableConsumerOpAuditModel(),
        )



import logging
from typing import Generator, List, Optional

from pydantic import conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGroupOp, BaseOp, BaseOpConfigModel
from nr_ops.ops.op import Op, OpModel

logger = logging.getLogger(__name__)


class OpSetGroupOpConfigModel(BaseOpConfigModel):
    ops: conlist(OpModel, min_items=1)

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    def get_self_and_child_op_ids(self) -> Generator[str, None, None]:
        """."""
        for op in self.ops:
            yield from op.get_self_and_child_op_ids()


class OpSetGroupOpMetadataModel(BaseOpMetadataModel):
    output_metadata: BaseOpMetadataModel

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OpSetGroupOpAuditModel(BaseOpAuditModel):
    output_audit: BaseOpAuditModel

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OpSetGroupOp(BaseGroupOp):
    OP_TYPE = "group.op_set"
    OP_CONFIG_MODEL = OpSetGroupOpConfigModel
    OP_METADATA_MODEL = OpSetGroupOpMetadataModel
    OP_AUDIT_MODEL = OpSetGroupOpAuditModel

    templated_fields = None

    def __init__(self, ops: List[BaseOp], **kwargs):
        self.ops = [Op(**op) for op in ops]

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[Optional[OpMsg], None, None]:
        """."""
        logger.info(f"OpSetGroupOp.run: Running")

        n_ops = len(self.ops)
        for op_i, op in enumerate(self.ops):
            # Exhaust each op generator.
            logger.info(
                f"OpSetGroupOp.run: Iterating over Ops, current Op "
                f"{op_i + 1}/{n_ops} (1-index) | {op.op_type=} | {op.op_id=}"
            )
            for output_msg_i, output_msg in enumerate(op.run(time_step, msg)):
                logger.info(
                    f"OpSetGroupOp.run: Consuming (not yielding) output_msg from Op "
                    f"{op_i + 1}/{n_ops} (1-index) | {op.op_type=} | {op.op_id=} | "
                    f"{output_msg_i} | {type(output_msg.data)=}"
                )
                pass

        # Yield original message to keep the chain going.
        logger.info(f"OpSetGroupOp.run: Yielding original input message {type(msg)=}.")
        yield msg



import logging
import time
from typing import Any, Dict, Literal

from google.cloud import bigquery
from pydantic import StrictStr, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel
from nr_ops.ops.ops.connector_ops.interfaces.base import validate_hook_type_and_config

logger = logging.getLogger(__name__)


class GCPBigQueryConnOpConfigModel(BaseOpConfigModel):
    hook_type: Literal[
        "connector.hooks.airflow_gcp_bigquery_hook",
    ]
    hook_config: Dict[StrictStr, Any]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    _validate_hook_type_and_config = root_validator(allow_reuse=True)(
        validate_hook_type_and_config
    )


class GCPBigQueryConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPBigQueryConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPBigQueryConnOp(BaseConnectorOp):
    OP_TYPE = "connector.gcp.bigquery"
    OP_CONFIG_MODEL = GCPBigQueryConnOpConfigModel
    OP_METADATA_MODEL = GCPBigQueryConnOpMetadataModel
    OP_AUDIT_MODEL = GCPBigQueryConnOpAuditModel

    templated_fields = None

    def __init__(self, hook_type: str, hook_config: Dict[str, Any], **kwargs):
        self.hook_type = hook_type
        self.hook_config = hook_config
        self.templated_fields = kwargs.get("templated_fields", [])

        from nr_ops.ops.op_collection import OP_COLLECTION

        self.hook_op = OP_COLLECTION[self.hook_type](**self.hook_config)
        self.hook_data = self.hook_op.run().data

        # Will be populated in run()
        self.bigquery_client: Any = None
        from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook

        self.bigquery_hook: BigQueryHook = None
        self.bigquery_client: bigquery.Client = None

    def extract_table(
        self,
        project: str,
        dataset_id: str,
        table_id: str,
        destination_uri: str,
        location: str,
        extract_job_config: Dict[str, Any],
    ):
        """."""
        logger.info(
            f"GCPBigQueryConnOp.extract_table: Extracting BigQuery table. "
            f"project={project}, dataset_id={dataset_id}, table_id={table_id}, "
            f"location={location}, destination_uri={destination_uri}"
        )
        start = time.time()
        dataset_ref = bigquery.DatasetReference(project, dataset_id)
        table_ref = dataset_ref.table(table_id)
        extract_job = self.bigquery_client.extract_table(
            table_ref,
            destination_uri,
            # Location must match that of the source table.
            location=location,
            job_config=bigquery.job.ExtractJobConfig(**extract_job_config),
        )
        extract_job.result()  # Waits for job to complete.
        end = time.time()
        logger.info(
            f"GCPBigQueryConnOp.extract_table: Done uploading BigQuery table "
            f"{dataset_id}.{table_id} to {destination_uri=}. "
            f"Time taken: {end - start:0.2f} seconds"
        )

    def get_tables(self, project: str, dataset_id: str):
        logger.info(
            f"GCPBigQueryConnOp.get_tables: Getting tables for "
            f"project={project}, dataset_id={dataset_id}"
        )
        tables = self.bigquery_hook.get_dataset_tables(
            project_id=project, dataset_id=dataset_id
        )
        logger.info(
            f"GCPBigQueryConnOp.get_tables: Done getting tables for "
            f"project={project}, dataset_id={dataset_id}. Fetched {len(tables)} tables"
        )

    def run(self) -> OpMsg:
        """."""
        logger.info(f"GCPBigQueryConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="GCPBigQueryConnOp.run:"
        )

        if self.hook_type == "connector.hooks.airflow_gcp_bigquery_hook":
            from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook

            self.hook_data: BigQueryHook

            self.bigquery_hook = self.hook_data
            self.bigquery_client = self.bigquery_hook.get_client()
        else:
            raise NotImplementedError()

        return OpMsg(
            data=self,
            metadata=GCPBigQueryConnOpMetadataModel(),
            audit=GCPBigQueryConnOpAuditModel(),
        )



import logging
from typing import Any, Dict, Generator, List, Literal, Optional

import pandas as pd
from pydantic import StrictBool, StrictStr, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpTimeStepMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseOpConfigModel, BaseTimeStepOp

logger = logging.getLogger(__name__)


class DateRangeTimeStepOpConfigModel(BaseOpConfigModel):
    start: StrictStr
    end: StrictStr
    tz: StrictStr
    freq: StrictStr
    inclusive: Literal["both", "neither", "left", "right"]
    metadata: Optional[Dict[str, Any]] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DateRangeTimeStepOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DateRangeTimeStepOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DateRangeTimeStepOp(BaseTimeStepOp):
    OP_TYPE = "ts.date_range"
    OP_CONFIG_MODEL = DateRangeTimeStepOpConfigModel
    OP_METADATA_MODEL = DateRangeTimeStepOpMetadataModel
    OP_AUDIT_MODEL = DateRangeTimeStepOpAuditModel

    templated_fields = None

    def __init__(
        self,
        start: str,
        end: str,
        freq: str,
        tz: str,
        inclusive: Literal["both", "neither", "left", "right"],
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        self.start = start
        self.end = end
        self.freq = freq
        self.tz = tz
        self.inclusive = inclusive
        self.metadata = {} if metadata is None else metadata

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(self) -> Generator[OpTimeStepMsg, None, None]:
        """."""
        logger.info(f"DateRangeTimeStepOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="DateRangeTimeStepOp.run:"
        )

        date_range = pd.date_range(
            start=self.start,
            end=self.end,
            freq=self.freq,
            tz=self.tz,
            inclusive=self.inclusive,
        )

        for start_date in date_range:
            time_step = TimeStep(
                start=start_date,
                end=start_date + pd.Timedelta(self.freq),
                tz=self.tz,
                metadata=self.metadata,
            )
            yield OpTimeStepMsg(
                data=time_step,
                metadata=DateRangeTimeStepOpMetadataModel(),
                audit=DateRangeTimeStepOpAuditModel(),
            )



def validate_hook_type_and_config(cls, values):
    """Validate that the config matches the type."""
    hook_type = values.get("hook_type")
    hook_config = values.get("hook_config")

    from nr_ops.ops.op_collection import OP_COLLECTION

    if hook_type not in OP_COLLECTION:
        raise ValueError(
            f"Hook with {hook_type=} was NOT FOUND in OP_COLLECTION. "
            f"Please check for typos in the `hook_type` or make sure the operator "
            f"class is correctly imported in OP_COLLECTION. FYI, a hook is also "
            f"an Connector Op that must exist in the OP_COLLECTION."
        )

    # Check if hook_config is a valid config for op_type.
    hook_model_cls = OP_COLLECTION[hook_type].OP_CONFIG_MODEL
    # hook_model is a pydantic model that is used to validate the hook_config.
    hook_model_cls(**hook_config)

    return values



import logging
import time
from typing import Literal, Union

from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.queue import QueueConnOp

logger = logging.getLogger(__name__)


class QueuePutConsumerOpConfigModel(BaseOpConfigModel):
    queue_conn_id: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class QueuePutConsumerOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class QueuePutConsumerOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class QueuePutConsumerOp(BaseConsumerOp):
    OP_TYPE = "consumer.queue.put"
    OP_CONFIG_MODEL = QueuePutConsumerOpConfigModel
    OP_METADATA_MODEL = QueuePutConsumerOpMetadataModel
    OP_AUDIT_MODEL = QueuePutConsumerOpAuditModel

    templated_fields = None

    def __init__(self, queue_conn_id: str, **kwargs):
        self.queue_conn_id = queue_conn_id

        op_manager = get_global_op_manager()

        self.queue_conn: Union[QueueConnOp] = op_manager.get_connector(
            op_id=self.queue_conn_id
        )

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(self, time_step: TimeStep, msg: OpMsg) -> OpMsg:
        """."""
        logger.info(f"QueuePutConsumerOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="QueuePutConsumerOp.run:"
        )

        fifo = self.queue_conn.get_reference()

        logger.info(
            f"QueuePutConsumerOp.run: Putting item of {type(msg.data)=} to queue"
        )
        fifo.put(msg.data)

        return OpMsg(
            data=None,
            metadata=QueuePutConsumerOpMetadataModel(),
            audit=QueuePutConsumerOpAuditModel(),
        )



import logging
from typing import Generator, List, Optional

from pydantic import conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGroupOp, BaseOp, BaseOpConfigModel
from nr_ops.ops.op import Op, OpModel

logger = logging.getLogger(__name__)


class OpFanInGroupOpConfigModel(BaseOpConfigModel):
    ops: conlist(OpModel, min_items=1)

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    def get_self_and_child_op_ids(self) -> Generator[str, None, None]:
        """."""
        for op in self.ops:
            yield from op.get_self_and_child_op_ids()


class OpFanInGroupOpMetadataModel(BaseOpMetadataModel):
    output_metadata: BaseOpMetadataModel

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OpFanInGroupOpAuditModel(BaseOpAuditModel):
    output_audit: BaseOpAuditModel

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OpFanInGroupOp(BaseGroupOp):
    """.

    This is a group op that yields the outputs of each op in the group as its output.
    - Think of it as a group of independent sources.
    - The input of this group op is passed to each op in the group.
    - Each op in the group is run to completion before the next op is run.
    - The output of each op is yielded out as the output of the group op AND not chained
      to other ops.
    """

    # Alternate names:
    # - group.group_of_generators
    # - group.group_of_sources

    OP_TYPE = "group.op_fan_in"
    OP_CONFIG_MODEL = OpFanInGroupOpConfigModel
    OP_METADATA_MODEL = OpFanInGroupOpMetadataModel
    OP_AUDIT_MODEL = OpFanInGroupOpAuditModel

    templated_fields = None

    def __init__(self, ops: List[BaseOp], **kwargs):
        self.ops = [Op(**op) for op in ops]

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[Optional[OpMsg], None, None]:
        """."""
        logger.info(f"OpFanInGroupOp.run: Running")

        n_ops = len(self.ops)
        for op_i, op in enumerate(self.ops):
            op_i += 1
            # Exhaust each op generator.
            # Pass the input message of this group op to each op.
            for output_msg_i, output_msg in enumerate(op.run(time_step, msg)):
                logger.info(
                    f"OpFanInGroupOp.run: Yielding output_msg from Op"
                    f"{op_i=}/{n_ops=} (1-index) | {op.op_type=} | {op.op_id=} | "
                    f"{output_msg_i} | {type(output_msg.data)=}"
                )

                # INHERIT metadata and audit from the input message.
                yield OpMsg(
                    data=output_msg.data,
                    metadata=output_msg.metadata,
                    audit=output_msg.audit,
                )



import copy
import importlib
import logging
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple, Union

import backoff
import requests
from pydantic import (
    BaseModel,
    Field,
    StrictBool,
    StrictStr,
    conlist,
    root_validator,
    validator,
)

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel
from nr_ops.ops.ops.connector_ops.hooks.http_requests_from_env import (
    HTTPRequestsHookFromEnvConnOp,
)
from nr_ops.ops.ops.connector_ops.interfaces.base import validate_hook_type_and_config

logger = logging.getLogger(__name__)


class BackoffExpoModel(BaseModel):
    base: float = 2
    factor: float = 2
    max_value: Optional[float] = None

    def wait_gen(self, *args, **kwargs) -> Generator[float, None, None]:
        return backoff.expo(
            base=self.base,
            factor=self.factor,
            max_value=self.max_value,
        )


class BackoffCustomValuesModel(BaseModel):
    values: conlist(Union[None, float], min_items=2) = Field(
        default_factory=lambda: [None, 5, 30, 60, 120, 180]
    )

    @validator("values")
    def validate_values(cls, values):
        if values[0] is not None:
            raise ValueError("First value must be None")

        for value in values[1:]:
            if value is None:
                raise ValueError(
                    "All values after the first value must be floats. "
                    "Only the first value must be None."
                )
        return values

    def wait_gen(self, *args, **kwargs) -> Generator[float, None, None]:
        for value in self.values:
            yield value


class BackoffConfigModel(BaseModel):
    exceptions: conlist(StrictStr, min_items=1) = Field(
        default_factory=lambda: [
            "requests.exceptions.RequestException",
            "urllib3.exceptions.HTTPError",
        ]
    )
    on_exception_kwargs: Dict[str, Any] = Field(
        default_factory=lambda: {"max_time": 400}
    )
    backoff_wait_gen_type: Literal["custom_values", "backoff.expo"] = "custom_values"
    backoff_wait_gen_config: Dict[str, Any] = Field(
        default_factory=lambda: BackoffCustomValuesModel().dict()
    )

    @root_validator
    def validate_backoff_wait_gen_config(cls, values):
        if values.get("backoff_wait_gen_type", None) == "custom_values":
            BackoffCustomValuesModel(**values["backoff_wait_gen_config"])
        elif values.get("backoff_wait_gen_type", None) == "backoff.expo":
            BackoffExpoModel(**values["backoff_wait_gen_config"])
        else:
            raise ValueError(
                f"Invalid backoff_wait_gen_type: "
                f"{values.get('backoff_wait_gen_type', None)}"
            )
        return values


class HTTPConnOpConfigModel(BaseOpConfigModel):
    hook_type: Literal[
        "connector.hooks.http_requests_from_env",
    ]
    hook_config: Dict[StrictStr, Any]

    # NOTE: If user does not specify backoff_config, it will be set to NONE and no
    # backoff config will be enabled. If user specifies backoff_config={} then
    # default backoff config will be enabled!
    backoff_config: Optional[BackoffConfigModel] = None

    # Example backoff_config:
    #   backoff_config:
    #     on_exception_kwargs:
    #       max_time: 400
    #     backoff_wait_gen_type: "custom_values"
    #     backoff_wait_gen_config:
    #       values: [null, 5, 30, 60, 120, 180]

    apply_basic_auth: StrictBool = False
    apply_extras_as_headers: StrictBool = True

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    _validate_hook_type_and_config = root_validator(allow_reuse=True)(
        validate_hook_type_and_config
    )


class HTTPConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class HTTPConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class HTTPConnOp(BaseConnectorOp):
    OP_TYPE = "connector.http"
    OP_CONFIG_MODEL = HTTPConnOpConfigModel
    OP_METADATA_MODEL = HTTPConnOpMetadataModel
    OP_AUDIT_MODEL = HTTPConnOpAuditModel

    templated_fields = None

    def __init__(
        self,
        hook_type: str,
        hook_config: Dict[str, Any],
        backoff_config: Optional[Dict[str, Any]] = None,
        apply_basic_auth: bool = False,
        apply_extras_as_headers: bool = True,
        **kwargs,
    ):
        self.hook_type = hook_type
        self.hook_config = hook_config
        # NOTE: If user does not specify backoff_config, it will be set to NONE and no
        # backoff config will be enabled. If user specifies backoff_config={} then
        # default backoff config will be enabled!
        self.backoff_config = backoff_config
        self.apply_basic_auth = apply_basic_auth
        self.apply_extras_as_headers = apply_extras_as_headers

        self.templated_fields = kwargs.get("templated_fields", [])

        from nr_ops.ops.op_collection import OP_COLLECTION

        self.hook_op = OP_COLLECTION[self.hook_type](**self.hook_config)
        self.hook = self.hook_op.run().data

        if self.hook_type == "connector.hooks.http_requests_from_env":
            self.hook: HTTPRequestsHookFromEnvConnOp
        else:
            raise NotImplementedError()

    def get_backoff_wrapper(self):
        """.

        * This function is called only if backoff_config is not None.
        """
        backoff_model = BackoffConfigModel(**self.backoff_config)
        exception_strs = (
            ["requests.exceptions.RequestException"]
            if backoff_model.exceptions is None
            else backoff_model.exceptions
        )
        exception_classes = []
        for exception_str in exception_strs:
            exception_split = exception_str.split(".")
            if len(exception_split) == 1:
                exception_module = "builtins"
                exception_name = exception_split[0]
            else:
                exception_module = ".".join(exception_split[:-1])
                exception_name = exception_split[-1]

            exception_class = getattr(
                importlib.import_module(exception_module), exception_name
            )
            exception_classes.append(exception_class)

        exception_classes = tuple(exception_classes)

        if backoff_model.backoff_wait_gen_type == "custom_values":
            wait_gen = BackoffCustomValuesModel(
                **backoff_model.backoff_wait_gen_config
            ).wait_gen
        elif backoff_model.backoff_wait_gen_type == "backoff.expo":
            wait_gen = BackoffExpoModel(
                **backoff_model.backoff_wait_gen_config
            ).wait_gen
        else:
            raise NotImplementedError()

        return backoff.on_exception(
            wait_gen=wait_gen,
            exception=exception_classes,
            jitter=None,
            **backoff_model.on_exception_kwargs,
        )

    ####################################################################################
    # PROPERTIES
    ####################################################################################
    @property
    def base_url(self) -> str:
        if self.hook_type == "connector.hooks.http_requests_from_env":
            # Base URL guaranteed to NOT have a trailing slash by hook.
            _base_url = self.hook.base_url
        else:
            raise NotImplementedError()

        return _base_url

    @property
    def username(self) -> str:
        if self.hook_type == "connector.hooks.http_requests_from_env":
            _username = self.hook.username
        else:
            raise NotImplementedError()

        return _username

    @property
    def password(self) -> str:
        if self.hook_type == "connector.hooks.http_requests_from_env":
            _password = self.hook.password
        else:
            raise NotImplementedError()

        return _password

    @property
    def extras(self) -> Dict[str, Any]:
        if self.hook_type == "connector.hooks.http_requests_from_env":
            _extras = self.hook.extras
        else:
            raise NotImplementedError()

        return _extras

    ####################################################################################
    # METHODS
    ####################################################################################

    def __call_logic(
        self,
        requests_method,
        url,
        return_type: Literal["json", "text"],
        requests_kwargs,
        accepted_status_codes: Optional[List[int]] = None,
    ):
        """."""
        if self.apply_basic_auth:
            logger.info(
                f"HTTPConnOp.__call_logic: `apply_basic_auth` is True. "
                f"Adding basic authentication from hook connection information."
                f"If auth are already present in requests_kwargs, then they will "
                f"take precedence over the auth from hook connection information."
            )
            requests_kwargs = copy.deepcopy(requests_kwargs)
            if not requests_kwargs.get("auth", None):
                requests_kwargs["auth"] = (self.username, self.password)

        if self.apply_extras_as_headers:
            requests_kwargs = copy.deepcopy(requests_kwargs)
            logger.info(
                f"HTTPConnOp.__call_logic: `apply_extras_as_headers` is True. "
                f"Adding extras as headers from hook connection information. "
                f"If headers are already present in requests_kwargs, then they will "
                f"take precedence over the headers from `extras`."
            )
            requests_kwargs["headers"] = {
                **self.extras,
                # Original requests kwargs headers take precedence over extras if exists
                **requests_kwargs.get("headers", {}),
            }

        response = requests_method(url, **requests_kwargs)

        if return_type == "json":
            status_code, response_data = response.status_code, response.json()
        elif return_type == "text":
            status_code, response_data = response.status_code, response.text
        else:
            raise NotImplementedError()

        if accepted_status_codes is None:
            logger.info(
                f"HTTPConnOp.__call_logic: No accepted_status_codes specified. "
                f"Not validating {status_code=}. Returning payload."
            )
            return status_code, response_data
        elif status_code in accepted_status_codes:
            logger.info(
                f"HTTPConnOp.__call_logic: {status_code=} is in "
                f"{accepted_status_codes=}. Returning payload."
            )
            return status_code, response_data
        else:
            msg = (
                f"HTTPConnOp.__call_logic: {accepted_status_codes=}, instead received "
                f"{status_code=}. Raising HTTPError so that it can be caught "
                f"by the backoff wrapper. response.text:\n{response.text}"
            )
            logger.exception(msg)
            raise requests.exceptions.HTTPError(msg)

    def call(
        self,
        method: Literal["get", "post", "put", "delete", "patch"],
        url: str,
        return_type: Literal["json", "text"],
        requests_kwargs: Optional[Dict[str, Any]] = None,
        accepted_status_codes: Optional[List[int]] = None,
    ) -> Tuple[int, Any]:
        """."""
        logger.info(f"HTTPConnOp.call: Running")

        logger.info(
            f"HTTPConnOp.call: Running for {method=}, {return_type=}, "
            f"{accepted_status_codes=}. Not logging URL since it may contain secrets."
        )
        requests_kwargs = {} if requests_kwargs is None else requests_kwargs

        requests_method = getattr(requests, method)

        # NOTE: If user does not specify backoff_config, it will be set to NONE and no
        # backoff config will be enabled. If user specifies backoff_config={} then
        # default backoff config will be enabled!
        if self.backoff_config is not None:
            logger.info(f"HTTPConnOp.call: Backoff enabled.")
            call_logic = self.get_backoff_wrapper()(self.__call_logic)
        else:
            logger.info(f"HTTPConnOp.call: Backoff NOT enabled.")
            call_logic = self.__call_logic

        return call_logic(
            requests_method=requests_method,
            url=url,
            return_type=return_type,
            requests_kwargs=requests_kwargs,
            accepted_status_codes=accepted_status_codes,
        )

    def run(self) -> OpMsg:
        """."""
        logger.info(f"HTTPConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(time_step=None, msg=None, log_prefix="HTTPConnOp.run:")

        if self.hook_type == "connector.hooks.http_requests_from_env":
            # Validate that if apply_basic_auth is True, then username and password
            # are not None.
            if self.apply_basic_auth:
                if (not self.username) or (not self.password):
                    raise ValueError(
                        f"HTTPConnOp.run: `apply_basic_auth` is True but username or "
                        f"password or both have not been provided."
                    )

        else:
            raise NotImplementedError()

        return OpMsg(
            data=self,
            metadata=HTTPConnOpMetadataModel(),
            audit=HTTPConnOpAuditModel(),
        )



"""."""
import logging

from numpy.ma import copy

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class PythonListHookConnOpConfigModel(BaseOpConfigModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonListHookConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonListHookConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonListHookConnOp(BaseConnectorOp):
    OP_TYPE = "connector.hooks.python.list"
    OP_CONFIG_MODEL = PythonListHookConnOpConfigModel
    OP_METADATA_MODEL = PythonListHookConnOpMetadataModel
    OP_AUDIT_MODEL = PythonListHookConnOpAuditModel

    templated_fields = None

    def __init__(self, **kwargs):
        self.templated_fields = kwargs.get("templated_fields", [])

        # NOTE: This is the list :)
        self._list = []

    def run(self) -> OpMsg:
        """."""
        logger.info(f"PythonListHookConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="PythonListHookConnOp.run:"
        )

        return OpMsg(
            data=self,
            metadata=PythonListHookConnOpMetadataModel(),
            audit=PythonListHookConnOpAuditModel(),
        )

    def get_reference(self):
        """."""
        return self._list

    def get_deepcopy(self):
        """."""
        return copy.deepcopy(self._list)

    def get_reference_and_reinit(self):
        """."""
        _list = self._list
        self._list = []
        return _list

    def put_append(self, item):
        """."""
        self._list.append(item)

    def put_extend(self, item):
        """."""
        self._list.extend(item)

    def size(self):
        """."""
        return len(self._list)



import logging
from typing import Literal, Union

from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.counter import CounterConnOp

logger = logging.getLogger(__name__)


class CounterPutConsumerOpConfigModel(BaseOpConfigModel):
    counter_conn_id: StrictStr
    put_type: Literal["update_dict", "update_list_of_dicts"]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class CounterPutConsumerOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class CounterPutConsumerOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class CounterPutConsumerOp(BaseConsumerOp):
    OP_TYPE = "consumer.counter.put"
    OP_CONFIG_MODEL = CounterPutConsumerOpConfigModel
    OP_METADATA_MODEL = CounterPutConsumerOpMetadataModel
    OP_AUDIT_MODEL = CounterPutConsumerOpAuditModel

    templated_fields = None

    def __init__(
        self,
        counter_conn_id: str,
        put_type: Literal["update_dict", "update_list_of_dicts"],
        **kwargs,
    ):
        self.counter_conn_id = counter_conn_id
        self.put_type = put_type

        op_manager = get_global_op_manager()

        self.counter_conn: Union[CounterConnOp] = op_manager.get_connector(
            op_id=self.counter_conn_id
        )

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(self, time_step: TimeStep, msg: OpMsg) -> OpMsg:
        """."""
        logger.info(f"CounterPutConsumerOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="CounterPutConsumerOp.run:"
        )

        logger.info(
            f"CounterPutConsumerOp.run: Putting item of {type(msg.data)=} "
            f"to counter with {self.put_type=}."
        )

        if self.put_type == "update_dict":
            self.counter_conn.put_update_dict(item=msg.data)
        elif self.put_type == "update_list_of_dicts":
            self.counter_conn.put_update_list_of_dicts(msg.data)
        else:
            raise NotImplementedError()

        logger.info(
            f"CounterPutConsumerOp.run: Done Putting item of {type(msg.data)=} "
            f"to counter with {self.put_type=}."
        )

        return OpMsg(
            data=None,
            metadata=CounterPutConsumerOpMetadataModel(),
            audit=CounterPutConsumerOpAuditModel(),
        )



from __future__ import annotations

import abc
import builtins
import copy
import json
import logging
from typing import Any, Dict, Generator, List, Literal, Optional, Type, Union

import jinja2
import jinja2.meta
from pydantic import BaseModel, StrictBool, StrictStr, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg, OpTimeStepMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.utils.eval.eval_globals import EVAL_GLOBALS

logger = logging.getLogger(__name__)


class RecursiveTemplateConfigModel(BaseModel):
    match_type: Literal["at_least_one", "all"]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class FieldTemplateConfigModel(BaseModel):
    field: StrictStr
    is_secret: StrictBool = False
    is_recursive: StrictBool = False
    recursive_config: Optional[RecursiveTemplateConfigModel] = None
    apply_json_serialization: StrictBool = False

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    @root_validator(pre=False)
    def validate_recursive_config(cls, values):
        if values.get("is_recursive") and not values.get("recursive_config"):
            raise ValueError("is_recursive is True, but no recursive_config provided")
        return values


class BaseOpConfigModel(BaseModel):
    """."""

    templated_fields: Optional[List[FieldTemplateConfigModel]] = None

    def get_self_and_child_op_ids(self) -> Generator[str, None, None]:
        """.

        NOTE: This method should be re-implemented by any Op that is from the
        `group` OP_FAMILY.
        """
        return
        yield

    @staticmethod
    def __is_templated_field_str(field: str, field_value: str) -> bool:
        """."""
        try:
            # source: https://jinja.palletsprojects.com/en/3.0.x/api/#jinja2.meta.find_undeclared_variables
            ast = jinja2.Environment().parse(source=field_value)
            vars_in_field_value = jinja2.meta.find_undeclared_variables(ast)
        except jinja2.exceptions.TemplateSyntaxError:
            logger.exception(
                f"BaseOpConfigModel.__is_templated_field_str: A TemplateSyntaxError "
                f"occurred when parsing: {field=}, {field_value=}. "
            )
            raise
        if len(vars_in_field_value) == 0:
            return False
        else:
            return True

    # A staticmethod that recursively loops through a list or dicts and calls the
    # __validate_templated_field_str method on each value that is of type str and calls
    # itself on each value that is of type list or dict.
    @staticmethod
    def __validate_templated_field_recursive(
        field: str,
        field_value: Union[List[Any], Dict[Any, Any]],
        n_templated_fields_found: int,
        n_templated_fields_validated: int,
    ) -> (int, int):
        """."""

        def _next_field_value_gen(field, field_value):
            if isinstance(field_value, list):
                for i, _value in enumerate(field_value):
                    yield f"{field}[{i}]", _value
            elif isinstance(field_value, dict):
                for key, _value in field_value.items():
                    yield f"{field}.{key}", _value
            elif isinstance(field_value, BaseModel):
                for key, _value in field_value.dict().items():
                    yield f"{field}.{key}", _value
            else:
                raise NotImplementedError()

        for next_field, next_value in _next_field_value_gen(field, field_value):
            if isinstance(next_value, str):
                is_validated = BaseOpConfigModel.__is_templated_field_str(
                    field=next_field, field_value=next_value
                )
                n_templated_fields_found += 1
                if is_validated:
                    n_templated_fields_validated += 1
            elif isinstance(next_value, (list, dict, BaseModel)):
                (
                    n_templated_fields_found,
                    n_templated_fields_validated,
                ) = BaseOpConfigModel.__validate_templated_field_recursive(
                    field=next_field,
                    field_value=next_value,
                    n_templated_fields_found=n_templated_fields_found,
                    n_templated_fields_validated=n_templated_fields_validated,
                )
            else:
                pass

        return (
            n_templated_fields_found,
            n_templated_fields_validated,
        )

    @root_validator(pre=False)
    def validate_templated_fields(cls, values):
        """."""
        templated_fields = values.get("templated_fields", None)
        if templated_fields is None:
            templated_fields = []

        for field_config in templated_fields:
            if isinstance(field_config, dict):
                field_config = FieldTemplateConfigModel(**field_config)
            elif isinstance(field_config, FieldTemplateConfigModel):
                pass
            else:
                raise NotImplementedError()

            field = field_config.field
            field_value = values.get(field, None)

            # In case we use nested pydantic models we need to convert them to dicts.
            # This is just for validation of strings, so we don't need to worry having
            # data types being messed up. Non string values will be skipped anyway.
            if isinstance(field_value, BaseModel):
                field_value = field_value.dict()

            if field_config.is_recursive:
                recursive_config = field_config.recursive_config
                if recursive_config is None:
                    raise ValueError(
                        f"validate_templated_fields: The `is_recursive` flag is set to "
                        f"True for the {field=} but `recursive_config` is None. A "
                        f"recursive_config is required when `is_recursive` is True."
                    )

                if isinstance(recursive_config, dict):
                    recursive_config = RecursiveTemplateConfigModel(**recursive_config)
                elif isinstance(recursive_config, RecursiveTemplateConfigModel):
                    pass
                else:
                    raise NotImplementedError()
            else:
                recursive_config = None

            if field == "templated_fields":
                raise ValueError(
                    f"validate_templated_fields: Cannot use `templated_fields` as a "
                    f"templated field. Please remove this from the config."
                )
            elif field not in values:
                raise ValueError(
                    f"validate_templated_fields: {field=} is not defined in the model "
                    f"but was added as one of templated_field. "
                    f"Please remove this from the config."
                )
            elif isinstance(field_value, str):
                is_validated = BaseOpConfigModel.__is_templated_field_str(
                    field=field, field_value=field_value
                )
                if not is_validated:
                    raise ValueError(
                        f"validate_templated_fields: {field=} is in the list of "
                        f"`templated_fields` but it does not have any templated "
                        f"variables inside its {field_value=}."
                    )
            elif isinstance(field_value, (list, dict)):
                if not field_config.is_recursive:
                    raise ValueError(
                        f"validate_templated_fields: {field=} has a field_value of "
                        f"type list or dict ({type(field_value)=}) but `is_recursive` "
                        f"flag is set to False. Please set `is_recursive` to True if "
                        f"you want to template this field_value."
                    )

                (
                    n_templated_fields_found,
                    n_templated_fields_validated,
                ) = BaseOpConfigModel.__validate_templated_field_recursive(
                    field=field,
                    field_value=field_value,
                    n_templated_fields_found=0,
                    n_templated_fields_validated=0,
                )

                if (recursive_config.match_type == "all") and (
                    n_templated_fields_validated != n_templated_fields_found
                ):
                    raise ValueError(
                        f"validate_templated_fields: {field=} is in the list of "
                        f"`templated_fields` with `is_recursive` set to True and "
                        f"{recursive_config.match_type=} and "
                        f"{n_templated_fields_validated=} and "
                        f"{n_templated_fields_found=} for the {field_value=}. Not ALL "
                        f"strings found in the templated field recursively are "
                        f"templated!"
                    )
                elif (recursive_config.match_type == "at_least_one") and (
                    n_templated_fields_validated == 0
                ):
                    raise ValueError(
                        f"validate_templated_fields: {field=} is in the list of "
                        f"`templated_fields` with `is_recursive` set to True and "
                        f"{recursive_config.match_type=} and "
                        f"{n_templated_fields_validated=} and "
                        f"{n_templated_fields_found=} for the {field_value=}. At least "
                        f"one string found in the templated field recursively are "
                        f"must be templated but None were found!"
                    )
                else:
                    pass
            else:
                raise NotImplementedError(
                    f"validate_templated_fields: {field=} of type {type(field_value)=} "
                    f"which is not supported."
                )

        return values


class BaseOp(abc.ABC):
    @property
    @classmethod
    @abc.abstractmethod
    def OP_FAMILY(cls) -> Literal["input_only", "output_only", "input_output"]:
        raise NotImplementedError()

    @property
    @classmethod
    @abc.abstractmethod
    def OP_TYPE(cls) -> str:
        raise NotImplementedError()

    @property
    @classmethod
    @abc.abstractmethod
    def OP_CONFIG_MODEL(cls) -> Type[BaseOpConfigModel]:
        raise NotImplementedError()

    @property
    @classmethod
    @abc.abstractmethod
    def OP_METADATA_MODEL(cls) -> Type[BaseOpMetadataModel]:
        raise NotImplementedError()

    @property
    @classmethod
    @abc.abstractmethod
    def OP_AUDIT_MODEL(cls) -> Type[BaseOpAuditModel]:
        raise NotImplementedError()

    # TODO: Should this be exposed to the user or should be take care of this in the
    #  background implicitly?
    # @property
    # @abc.abstractmethod
    # def raw_fields(self) -> Optional[Any]:
    #     raise NotImplementedError()

    @property
    @abc.abstractmethod
    def templated_fields(self) -> Optional[Any]:
        raise NotImplementedError()

    def __render_field_str(
        self,
        field: str,
        field_value: str,
        field_config: Any,
        op_manager: Any,
        rendered_fields: Any,
        time_step: Optional[TimeStep],
        msg: Optional[OpMsg],
        log_prefix: str = "",
    ):
        """."""
        ast = jinja2.Environment().parse(source=field_value)
        vars_in_field_value = jinja2.meta.find_undeclared_variables(ast)

        if len(vars_in_field_value) == 0:
            logger.info(f"{log_prefix} NOT Rendering field, no vars found: {field=}")
            return field_value
        else:
            pass

        logger.info(f"{log_prefix} Rendering field: {field=}")
        rendered_field_value = jinja2.Template(field_value).render(
            **{
                "builtins": builtins,
                "this_op": self.__dict__,
                "op_manager": op_manager,
                "time_step": time_step,
                "msg": msg,
                "rendered_fields": rendered_fields,
                **EVAL_GLOBALS,
            }
        )

        if field_config.is_secret:
            logger.info(f"{log_prefix} Rendered {field=}: rendered_field_value=******")
        else:
            logger.info(f"{log_prefix} Rendered {field=}: {rendered_field_value=}")

        if field_config.apply_json_serialization:
            logger.info(
                f"{log_prefix} Rendered {field=}: "
                f"{field_config.apply_json_serialization=}. "
                f"Applying json.loads on rendered_field_value."
            )
            rendered_field_value = json.loads(rendered_field_value)
            logger.info(
                f"{log_prefix} Rendered {field=}: "
                f"{field_config.apply_json_serialization=}. "
                f"Done serializing rendered_field_value with json.loads. "
            )

            # log rendered_field_value again, this time with types
            if field_config.is_secret:
                logger.info(
                    f"{log_prefix} Serialized {field=}: rendered_field_value=****** | "
                    f"{type(rendered_field_value)=}"
                )
            else:
                logger.info(
                    f"{log_prefix} Serialized {field=}: {rendered_field_value=} | "
                    f"{type(rendered_field_value)=}"
                )

        return rendered_field_value

    def __render_field_recursive(
        self,
        field: str,
        field_value: Any,
        field_config: Any,
        op_manager: Any,
        rendered_fields: Any,
        time_step: Optional[TimeStep],
        msg: Optional[OpMsg],
        log_prefix: str = "",
    ):
        """."""
        if isinstance(field_value, list):
            rendered_list = [
                self.__render_field_recursive(
                    field=f"{field}[{i}]",
                    field_value=next_field_value,
                    field_config=field_config,
                    op_manager=op_manager,
                    rendered_fields=rendered_fields,
                    time_step=time_step,
                    msg=msg,
                    log_prefix=log_prefix,
                )
                for i, next_field_value in enumerate(field_value)
            ]
            return rendered_list

        elif isinstance(field_value, dict):
            rendered_dict = {
                key: self.__render_field_recursive(
                    field=f"{field}.{key}",
                    field_value=next_field_value,
                    field_config=field_config,
                    op_manager=op_manager,
                    rendered_fields=rendered_fields,
                    time_step=time_step,
                    msg=msg,
                    log_prefix=log_prefix,
                )
                for key, next_field_value in field_value.items()
            }
            return rendered_dict

        elif isinstance(field_value, BaseModel):
            rendered_model = type(field_value)(
                **{
                    key: self.__render_field_recursive(
                        field=f"{field}.{key}",
                        field_value=next_field_value,
                        field_config=field_config,
                        op_manager=op_manager,
                        rendered_fields=rendered_fields,
                        time_step=time_step,
                        msg=msg,
                        log_prefix=log_prefix,
                    )
                    for key, next_field_value in field_value.dict().items()
                }
            )
            return rendered_model

        elif isinstance(field_value, str):
            return self.__render_field_str(
                field=field,
                field_value=field_value,
                field_config=field_config,
                op_manager=op_manager,
                rendered_fields=rendered_fields,
                time_step=time_step,
                msg=msg,
                log_prefix=log_prefix,
            )
        else:
            return field_value

    def render_fields(
        self, time_step: Optional[TimeStep], msg: Optional[OpMsg], log_prefix: str = ""
    ) -> Dict[str, str]:
        """Render all fields from templated fields.

        The templates are rendered using the `self` object as the context i.e.
        the templates are the attributes of the `self` object which are defined in the
        templated_fields list.

        We rerender the fields for every time_step and msg. This is because the
        rendered fields may depend on the time_step and msg.

        We store the original raw fields (as provided by the user) in the raw_fields
        attribute. This is so that we can re-render the fields if the time_step or msg
        changes.
        """
        logger.info(f"{log_prefix} Rendering all templated_fields")

        if self.templated_fields is None:
            raise ValueError(
                f"{log_prefix}: {self.OP_TYPE=} has not handled templated_fields in "
                f"its __init__ method."
            )

        # TODO: Should the raw_fields be exposed to the user or is the current behavior
        #  of taking care of it implicitly in the background okay?
        # Initialize raw fields
        if not hasattr(self, "raw_fields"):
            self.raw_fields = {}
            for field_config in self.templated_fields:
                field_config = FieldTemplateConfigModel(**field_config)
                field = field_config.field
                field_value = getattr(self, field)
                # NOTE: Absolutely necessary to use deepcopy here.
                # NOTE: We are *almost* guaranteed that only JSON serializable types
                # are used in the templated fields.
                self.raw_fields[field] = copy.deepcopy(field_value)

        from nr_ops.ops.op_manager import get_global_op_manager

        op_manager = get_global_op_manager()
        rendered_fields = {}
        for field_config in self.templated_fields:
            field_config = FieldTemplateConfigModel(**field_config)
            if field_config.is_recursive:
                field = field_config.field
                # NOTE: Absolutely necessary to use deepcopy here.
                # Use raw_fields[field] instead of getattr(self, field)
                field_value = copy.deepcopy(self.raw_fields[field])
                rendered_fields[field] = self.__render_field_recursive(
                    field=field,
                    field_value=field_value,
                    field_config=field_config,
                    op_manager=op_manager,
                    rendered_fields=rendered_fields,
                    time_step=time_step,
                    msg=msg,
                    log_prefix=log_prefix,
                )
            else:
                field = field_config.field
                # NOTE: Absolutely necessary to use deepcopy here.
                # Use raw_fields[field] instead of getattr(self, field)
                field_value = copy.deepcopy(self.raw_fields[field])
                rendered_fields[field] = self.__render_field_str(
                    field=field,
                    field_value=field_value,
                    field_config=field_config,
                    op_manager=op_manager,
                    rendered_fields=rendered_fields,
                    time_step=time_step,
                    msg=msg,
                    log_prefix=log_prefix,
                )

        # ------------------------------------------------------------------------------
        # WARNING: THIS IS A MAJOR HACK/FEATURE THAT CAN HAVE UNEXPECTED SIDE EFFECTS
        # ------------------------------------------------------------------------------
        for field, rendered_field_value in rendered_fields.items():
            setattr(self, field, rendered_field_value)

        logger.info(f"{log_prefix} Done rendering all templated_fields")

        return rendered_fields


class BaseGeneratorOp(BaseOp, abc.ABC):
    OP_FAMILY: Literal["generator"] = "generator"

    @abc.abstractmethod
    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        raise NotImplementedError()


class BaseConsumerOp(BaseOp, abc.ABC):
    OP_FAMILY: Literal["consumer"] = "consumer"

    @abc.abstractmethod
    def run(self, time_step: TimeStep, msg: OpMsg) -> OpMsg:  # Return not Yield
        raise NotImplementedError()


class BaseGroupOp(BaseOp, abc.ABC):
    OP_FAMILY: Literal["group"] = "group"

    @abc.abstractmethod
    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[Optional[OpMsg], None, None]:
        raise NotImplementedError()


class BaseTimeStepOp(BaseOp, abc.ABC):
    OP_FAMILY: Literal["time_step"] = "time_step"

    @abc.abstractmethod
    def run(self) -> Generator[OpTimeStepMsg, None, None]:
        raise NotImplementedError()


class BaseConnectorOp(BaseOp, abc.ABC):
    OP_FAMILY: Literal["connector"] = "connector"

    @abc.abstractmethod
    def run(self) -> OpMsg:  # Return not Yield
        raise NotImplementedError()



import collections
import logging
from typing import Dict, Iterable, List, Type

from nr_ops.ops.base import BaseOp
from nr_ops.ops.ops.connector_ops.hooks.airflow_gcp_bigquery_hook import (
    AirflowGCPBigQueryHookConnOp,
)
from nr_ops.ops.ops.connector_ops.hooks.airflow_gcp_gcs_hook import (
    AirflowGCPGCSHookConnOp,
)
from nr_ops.ops.ops.connector_ops.hooks.airflow_gcp_hook import AirflowGCPHookConnOp
from nr_ops.ops.ops.connector_ops.hooks.airflow_mysql_hook import AirflowMysqlHookConnOp
from nr_ops.ops.ops.connector_ops.hooks.airflow_postgres_hook import (
    AirflowPostgresHookConnOp,
)
from nr_ops.ops.ops.connector_ops.hooks.airflow_s3_hook import AirflowS3HookConnOp
from nr_ops.ops.ops.connector_ops.hooks.gcp_service_account_from_file import (
    GCPServiceAccountFromFileConnOp,
)
from nr_ops.ops.ops.connector_ops.hooks.http_requests_from_env import (
    HTTPRequestsHookFromEnvConnOp,
)
from nr_ops.ops.ops.connector_ops.hooks.python_counter import PythonCounterHookConnOp
from nr_ops.ops.ops.connector_ops.hooks.python_file import PythonFileHookConnOp
from nr_ops.ops.ops.connector_ops.hooks.python_list import PythonListHookConnOp
from nr_ops.ops.ops.connector_ops.hooks.python_queue import PythonQueueHookConnOp
from nr_ops.ops.ops.connector_ops.interfaces.counter import CounterConnOp
from nr_ops.ops.ops.connector_ops.interfaces.file import FileConnOp
from nr_ops.ops.ops.connector_ops.interfaces.gcp_bigquery import GCPBigQueryConnOp
from nr_ops.ops.ops.connector_ops.interfaces.gcp_gcs import GCPGCSConnOp
from nr_ops.ops.ops.connector_ops.interfaces.google_ads import GoogleAdsConnectorOp
from nr_ops.ops.ops.connector_ops.interfaces.google_analytics import (
    GoogleAnalyticsConnOp,
)
from nr_ops.ops.ops.connector_ops.interfaces.google_analytics_ga4 import (
    GoogleAnalyticsGA4ConnOp,
)
from nr_ops.ops.ops.connector_ops.interfaces.google_sheets import GoogleSheetsConnOp
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp
from nr_ops.ops.ops.connector_ops.interfaces.list import ListConnOp
from nr_ops.ops.ops.connector_ops.interfaces.mysql import MysqlConnOp
from nr_ops.ops.ops.connector_ops.interfaces.postgres import PostgresConnOp
from nr_ops.ops.ops.connector_ops.interfaces.queue import QueueConnOp
from nr_ops.ops.ops.connector_ops.interfaces.s3 import S3ConnOp
from nr_ops.ops.ops.consumer_ops.braze.generic_post_json_data_batched_async import (
    BrazeGenericPostJsonDataAsyncOp,
)
from nr_ops.ops.ops.consumer_ops.counter.counter_put import CounterPutConsumerOp
from nr_ops.ops.ops.consumer_ops.dbt.dbt_run import DBTRunConsumerOp
from nr_ops.ops.ops.consumer_ops.file.file_put import FilePutConsumerOp
from nr_ops.ops.ops.consumer_ops.gcp_bigquery.extract_table import GCPBigQueryToGCSOp
from nr_ops.ops.ops.consumer_ops.gcp_gcs.delete_key import GCPGCSDeleteKeyOp
from nr_ops.ops.ops.consumer_ops.gcp_gcs.put_key import GCPGCSPutKeyOp
from nr_ops.ops.ops.consumer_ops.google_ads.upload_offline_conversion import (
    GoogleAdsUploadOfflineConversionOp,
)
from nr_ops.ops.ops.consumer_ops.list.list_put import ListPutConsumerOp
from nr_ops.ops.ops.consumer_ops.mock import MockConsumerOp
from nr_ops.ops.ops.consumer_ops.nr_ops.postgres_create_etl_table import (
    NROpsPostgresCreateETLTableConsumerOp,
)
from nr_ops.ops.ops.consumer_ops.pangres.df_to_sql_db import PangresDFToSQLDBOp
from nr_ops.ops.ops.consumer_ops.put_list import PutListConsumerOp
from nr_ops.ops.ops.consumer_ops.queue.queue_put import QueuePutConsumerOp
from nr_ops.ops.ops.consumer_ops.s3.put_key import S3PutKeyOp
from nr_ops.ops.ops.consumer_ops.shell.shell_run import ShellRunConsumerOp
from nr_ops.ops.ops.consumer_ops.sql_query import SQLQueryConsumerOp
from nr_ops.ops.ops.generator_ops.airflow.dag.clear_task_instances import (
    AirflowDagClearTaskInstancesOp,
)
from nr_ops.ops.ops.generator_ops.airflow.dagruns.clear_dagrun import (
    AirflowDagRunClearDagRunOp,
)
from nr_ops.ops.ops.generator_ops.airflow.dagruns.get_dagrun import (
    AirflowDagRunGetDagRunOp,
)
from nr_ops.ops.ops.generator_ops.airflow.dagruns.trigger_dagrun import (
    AirflowDagRunTriggerDagRunOp,
)
from nr_ops.ops.ops.generator_ops.airflow.dagruns.trigger_dagrun_and_wait_until_completion import (
    AirflowDagRunTriggerDagAndWaitUntilCompletionOp,
)
from nr_ops.ops.ops.generator_ops.batch import BatchGeneratorOp
from nr_ops.ops.ops.generator_ops.bigcom.catalog_get_all_brands import (
    BigComCatalogGetAllBrandsOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.catalog_get_all_categories import (
    BigComCatalogGetAllCategoriesOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.coupons_get_all_coupons import (
    BigComCouponsGetAllCouponsOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.customers_get_all_customers import (
    BigComCustomersGetAllCustomersOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.generic_compute_email_hashes import (
    BigcomGenericComputeEmailHashesOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.generic_remove_pii import (
    BigcomGenericRemovePiiOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.orders_get_all_order_coupons import (
    BigComOrdersGetAllOrderCouponsOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.orders_get_all_order_products import (
    BigComOrdersGetAllOrderProductsOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.orders_get_all_order_shipping_addresses import (
    BigComOrdersGetAllOrderShippingAddressesOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.orders_get_all_orders import (
    BigComOrdersGetAllOrdersOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.products_get_all_product_variants import (
    BigComProductsGetAllProductVariantsOp,
)
from nr_ops.ops.ops.generator_ops.bigcom.products_get_all_products import (
    BigComProductsGetAllProductsOp,
)
from nr_ops.ops.ops.generator_ops.blade.get_token import BladeGetTokenOp
from nr_ops.ops.ops.generator_ops.blade.orders_list_goodsout import (
    BladeOrdersListGoodsoutOp,
)
from nr_ops.ops.ops.generator_ops.blade.orders_list_metadata import (
    BladeOrdersListMetadataOp,
)
from nr_ops.ops.ops.generator_ops.blade.orders_list_tracking_numbers import (
    BladeOrdersBulkListTrackingNumbersOp,
)
from nr_ops.ops.ops.generator_ops.blade.orders_view_goodsout import (
    BladeOrdersViewGoodsoutOp,
)
from nr_ops.ops.ops.generator_ops.blade.products_list_variations import (
    BladeProductsListVariationsOp,
)
from nr_ops.ops.ops.generator_ops.blade.products_view_variation import (
    BladeProductsViewVariationOp,
)
from nr_ops.ops.ops.generator_ops.compression.compress import CompressOp
from nr_ops.ops.ops.generator_ops.compression.decompress import DecompressOp
from nr_ops.ops.ops.generator_ops.compression.decompress_utils.decompress_newline_separated_streams_with_zlib import (
    DecompressNewlineSeparatedStreamsWithZlibOp,
)
from nr_ops.ops.ops.generator_ops.eval_expr.eval_expr import EvalExprOp
from nr_ops.ops.ops.generator_ops.eval_expr.eval_expr_as_metadata import (
    EvalExprAsMetadataOp,
)
from nr_ops.ops.ops.generator_ops.eval_expr.eval_expr_conditional import (
    EvalExprConditionalOp,
)
from nr_ops.ops.ops.generator_ops.gcp_gcs.get_key import GCPGCSGetKeyOp
from nr_ops.ops.ops.generator_ops.gcp_gcs.is_key_exists import GCPGCSIsKeyExistsOp
from nr_ops.ops.ops.generator_ops.gcp_gcs.list_keys import GCPGCSListKeysOp
from nr_ops.ops.ops.generator_ops.google.get_ga_reports import GetGAReportsOp
from nr_ops.ops.ops.generator_ops.google.get_ga_reports_ga4 import GetGAReportsGA4Op
from nr_ops.ops.ops.generator_ops.google.google_sheets_get_cell_range import (
    GoogleSheetsGetCellRangeOp,
)
from nr_ops.ops.ops.generator_ops.google_ads.google_ads_search_query import (
    GoogleAdsSearchQueryGeneratorOp,
)
from nr_ops.ops.ops.generator_ops.list.list_get import ListGetGeneratorOp
from nr_ops.ops.ops.generator_ops.mailchimp.campaigns_get_campaign_info import (
    MailchimpCampaignsGetCampaignInfoOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.campaigns_get_campaign_reports import (
    MailchimpCampaignsGetCampaignReportsOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.campaigns_get_campaigns import (
    MailchimpCampaignsGetCampaignsOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.campaigns_get_email_activity import (
    MailchimpCampaignsGetEmailActivityOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.campaigns_get_links_clicked import (
    MailchimpCampaignsGetLinksClickedOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.campaigns_get_links_clicked_members import (
    MailchimpCampaignsGetLinksClickedMembersOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.campaigns_get_sent_to import (
    MailchimpCampaignsGetSentToOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.campaigns_get_unsubscribed_members import (
    MailchimpCampaignsGetUnsubscribedMembersOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.generic_compute_email_hashes import (
    MailchimpGenericComputeEmailHashesOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.generic_remove_links import (
    MailchimpGenericRemoveLinksOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.generic_remove_pii import (
    MailchimpGenericRemovePiiOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.lists_get_list_members import (
    MailchimpListsGetListMembersOp,
)
from nr_ops.ops.ops.generator_ops.mailchimp.lists_get_lists import (
    MailchimpListsGetListsOp,
)
from nr_ops.ops.ops.generator_ops.mock import MockGeneratorOp
from nr_ops.ops.ops.generator_ops.ongoing.generic_remove_pii import (
    OngoingGenericRemovePiiOp,
)
from nr_ops.ops.ops.generator_ops.ongoing.orders_rest_get_all_orders import (
    OngoingOrdersRESTGetAllOrdersOp,
)
from nr_ops.ops.ops.generator_ops.ongoing.orders_rest_get_order import (
    OngoingOrdersRESTGetOrderOp,
)
from nr_ops.ops.ops.generator_ops.ongoing.orders_rest_get_waybill_rows import (
    OngoingOrdersRESTGetWayBillRowsOp,
)
from nr_ops.ops.ops.generator_ops.ongoing.orders_soap_get_all_orders import (
    OngoingOrdersSOAPGetAllOrdersOp,
)
from nr_ops.ops.ops.generator_ops.pandas.read_generic import PandasReadGenericOp
from nr_ops.ops.ops.generator_ops.pandas.train_and_test_split import (
    PandasTrainTestSplit,
)
from nr_ops.ops.ops.generator_ops.pickle.pickle import PickleGeneratorOp
from nr_ops.ops.ops.generator_ops.pickle.unpickle import UnPickleGeneratorOp
from nr_ops.ops.ops.generator_ops.priceindx.download_export_pipeline import (
    PriceindxDownloadExportPipelineOp,
)
from nr_ops.ops.ops.generator_ops.queue.queue_get import QueueGetGeneratorOp
from nr_ops.ops.ops.generator_ops.s3.get_key import S3GetKeyOp
from nr_ops.ops.ops.generator_ops.s3.list_keys import S3ListKeysOp
from nr_ops.ops.ops.generator_ops.skip_empty_input import SkipEmptyInputGeneratorOp
from nr_ops.ops.ops.generator_ops.sleep import SleepGeneratorOp
from nr_ops.ops.ops.generator_ops.sql_query import SQLQueryGeneratorOp
from nr_ops.ops.ops.group_ops.op_chain import OpChainGroupOp
from nr_ops.ops.ops.group_ops.op_chain_branch import OpChainBranchGroupOp
from nr_ops.ops.ops.group_ops.op_fan_in import OpFanInGroupOp
from nr_ops.ops.ops.group_ops.op_set import OpSetGroupOp
from nr_ops.ops.ops.time_step_ops.date_range import DateRangeTimeStepOp
from nr_ops.ops.ops.time_step_ops.mock_ts import MockTimeStepOp
from nr_ops.ops.ops.time_step_ops.simple_start_offset import SimpleStartOffsetTimeStepOp
from nr_ops.ops.ops.time_step_ops.simple_ts import SimpleTimeStepOp

logger = logging.getLogger(__name__)


OP_CLASSES: List[Type[BaseOp]] = [
    # connector_ops (Hooks)
    AirflowGCPHookConnOp,
    AirflowGCPBigQueryHookConnOp,
    AirflowGCPGCSHookConnOp,
    AirflowPostgresHookConnOp,
    AirflowMysqlHookConnOp,
    GCPServiceAccountFromFileConnOp,
    AirflowS3HookConnOp,
    HTTPRequestsHookFromEnvConnOp,
    PythonFileHookConnOp,
    PythonListHookConnOp,
    PythonCounterHookConnOp,
    PythonQueueHookConnOp,
    # connector_ops (Interfaces)
    PostgresConnOp,
    MysqlConnOp,
    S3ConnOp,
    HTTPConnOp,
    FileConnOp,
    ListConnOp,
    CounterConnOp,
    QueueConnOp,
    GCPGCSConnOp,
    GCPBigQueryConnOp,
    GoogleAdsConnectorOp,
    GoogleAnalyticsConnOp,
    GoogleAnalyticsGA4ConnOp,
    GoogleSheetsConnOp,
    # consumer_ops
    MockConsumerOp,
    PangresDFToSQLDBOp,
    SQLQueryConsumerOp,
    S3PutKeyOp,
    PutListConsumerOp,
    GCPBigQueryToGCSOp,
    GCPGCSPutKeyOp,
    GCPGCSDeleteKeyOp,
    GoogleAdsUploadOfflineConversionOp,
    ShellRunConsumerOp,
    DBTRunConsumerOp,
    FilePutConsumerOp,
    ListPutConsumerOp,
    QueuePutConsumerOp,
    CounterPutConsumerOp,
    NROpsPostgresCreateETLTableConsumerOp,
    BrazeGenericPostJsonDataAsyncOp,
    # generator_ops
    AirflowDagRunTriggerDagRunOp,
    AirflowDagRunGetDagRunOp,
    AirflowDagRunClearDagRunOp,
    AirflowDagClearTaskInstancesOp,
    AirflowDagRunTriggerDagAndWaitUntilCompletionOp,
    EvalExprOp,
    EvalExprAsMetadataOp,
    EvalExprConditionalOp,
    MockGeneratorOp,
    SleepGeneratorOp,
    SkipEmptyInputGeneratorOp,
    S3ListKeysOp,
    S3GetKeyOp,
    CompressOp,
    DecompressOp,
    DecompressNewlineSeparatedStreamsWithZlibOp,
    BladeGetTokenOp,
    BladeOrdersListGoodsoutOp,
    BladeOrdersViewGoodsoutOp,
    BladeOrdersListMetadataOp,
    BladeOrdersBulkListTrackingNumbersOp,
    BladeProductsListVariationsOp,
    BladeProductsViewVariationOp,
    ListGetGeneratorOp,
    QueueGetGeneratorOp,
    BigComOrdersGetAllOrdersOp,
    BigComOrdersGetAllOrderProductsOp,
    BigComOrdersGetAllOrderCouponsOp,
    BigComOrdersGetAllOrderShippingAddressesOp,
    BigComProductsGetAllProductsOp,
    BigComProductsGetAllProductVariantsOp,
    BigComCustomersGetAllCustomersOp,
    BigComCatalogGetAllBrandsOp,
    BigComCatalogGetAllCategoriesOp,
    BigComCouponsGetAllCouponsOp,
    BigcomGenericComputeEmailHashesOp,
    BigcomGenericRemovePiiOp,
    MailchimpCampaignsGetCampaignsOp,
    MailchimpCampaignsGetCampaignInfoOp,
    MailchimpCampaignsGetSentToOp,
    MailchimpCampaignsGetUnsubscribedMembersOp,
    MailchimpCampaignsGetLinksClickedOp,
    MailchimpCampaignsGetLinksClickedMembersOp,
    MailchimpListsGetListsOp,
    MailchimpListsGetListMembersOp,
    MailchimpCampaignsGetCampaignReportsOp,
    MailchimpCampaignsGetEmailActivityOp,
    MailchimpGenericComputeEmailHashesOp,
    MailchimpGenericRemovePiiOp,
    MailchimpGenericRemoveLinksOp,
    OngoingOrdersSOAPGetAllOrdersOp,
    OngoingOrdersRESTGetWayBillRowsOp,
    OngoingOrdersRESTGetAllOrdersOp,
    OngoingOrdersRESTGetOrderOp,
    OngoingGenericRemovePiiOp,
    PriceindxDownloadExportPipelineOp,
    PandasReadGenericOp,
    PandasTrainTestSplit,
    GCPGCSGetKeyOp,
    GCPGCSListKeysOp,
    GCPGCSIsKeyExistsOp,
    GetGAReportsOp,
    GetGAReportsGA4Op,
    GoogleAdsSearchQueryGeneratorOp,
    GoogleSheetsGetCellRangeOp,
    BatchGeneratorOp,
    SQLQueryGeneratorOp,
    PickleGeneratorOp,
    UnPickleGeneratorOp,
    # group_ops (Groups)
    OpChainGroupOp,
    OpSetGroupOp,
    OpFanInGroupOp,
    OpChainBranchGroupOp,
    # time_step_ops (TimeSteps)
    MockTimeStepOp,
    SimpleTimeStepOp,
    SimpleStartOffsetTimeStepOp,
    DateRangeTimeStepOp,
]


def validate_duplicate_op_types(op_classes: Iterable[Type[BaseOp]]):
    """."""
    op_type_counts = collections.Counter([op_class.OP_TYPE for op_class in op_classes])
    duplicate_op_types = [
        op_type for op_type, count in op_type_counts.items() if count > 1
    ]
    if duplicate_op_types:
        raise ValueError(
            f"validate_duplicate_op_types: The following op_types have been duplicated "
            f"{duplicate_op_types=}. Please ensure that all op_types are unique."
        )


validate_duplicate_op_types(OP_CLASSES)


OP_COLLECTION: Dict[str, Type[BaseOp]] = {}  # type: ignore

for op_class in OP_CLASSES:
    OP_COLLECTION[op_class.OP_TYPE] = op_class  # type: ignore

# ----------------------------------------------------------------------------------
# ALIASES
# ----------------------------------------------------------------------------------
ALIAS_COLLECTION: Dict[str, Type[BaseOp]] = {
    "group.op_list": OpChainGroupOp,
}

for alias, op_class in ALIAS_COLLECTION.items():
    if alias in OP_COLLECTION:
        raise ValueError(
            f"ALIAS_COLLECTION: The {alias=} is already in use. Please choose "
            f"a different alias."
        )
    OP_COLLECTION[alias] = op_class



import logging
import time
from collections import Counter
from typing import Any, Dict, Generator, Optional, Union

import psutil
from pydantic import BaseModel, StrictBool, StrictStr, root_validator

from nr_ops.messages.op_msg import OpMsg, OpTimeStepMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import (
    BaseConnectorOp,
    BaseConsumerOp,
    BaseGeneratorOp,
    BaseGroupOp,
    BaseTimeStepOp,
)

logger = logging.getLogger(__name__)


class OpModel(BaseModel):
    op_id: Optional[StrictStr] = None
    op_docs: Optional[StrictStr] = None
    op_type: StrictStr
    op_config: Dict[StrictStr, Any]
    store_metadata: StrictBool = True
    store_data: StrictBool = False

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    @root_validator(pre=False)
    def validate_op_type_and_config(cls, values):
        """Validate that the config matches the type."""
        op_type = values.get("op_type")
        op_config = values.get("op_config")

        from .op_collection import OP_COLLECTION

        if op_type not in OP_COLLECTION:
            raise ValueError(
                f"Operator with {op_type=} was NOT FOUND in OP_COLLECTION. "
                f"Please check for typos in the `op_type` or make sure the operator "
                f"class is correctly imported in OP_COLLECTION."
            )

        # Check if op_config is a valid config for op_type.
        op_model_cls = OP_COLLECTION[op_type].OP_CONFIG_MODEL
        # op_model is a pydantic model that is used to validate the op_config.
        op_model_cls(**op_config)

        return values

    def get_self_and_child_op_ids(self) -> Generator[str, None, None]:
        """.

        NOTE: This method is called once from main on only the root op.
        - The root op will call this method on all of its children.
        """
        if self.op_id is not None:
            yield self.op_id

        from .op_collection import OP_COLLECTION

        op_model_cls = OP_COLLECTION[self.op_type].OP_CONFIG_MODEL
        op_model = op_model_cls(**self.op_config)

        yield from op_model.get_self_and_child_op_ids()

    def validate_unique_ids(self):
        """.

        NOTE: This method is called once from main on only the root op.
        - The root op will call this method on all of its children.
        """

        all_ids = list(self.get_self_and_child_op_ids())
        counts = Counter(all_ids)
        duplicates = [op_id for op_id, count in counts.items() if count > 1]
        if duplicates:
            raise ValueError(
                f"Duplicate op_ids found: {duplicates}. op_ids must be unique across "
                f"all operators."
            )


class Op(object):
    def __init__(
        self,
        op_type: str,
        op_config: Dict[str, Any],
        op_id: Optional[str] = None,
        op_docs: Optional[str] = None,
        store_metadata: bool = True,
        store_data: bool = False,
    ):
        self.op_id = op_id
        self.op_docs = op_docs
        self.op_type = op_type
        self.op_config = op_config
        self.store_metadata = store_metadata
        self.store_data = store_data

        from .op_collection import OP_COLLECTION

        self.op_class = OP_COLLECTION[self.op_type]
        self.op_obj = self.op_class(**self.op_config)

        from nr_ops.ops.op_manager import get_global_op_manager

        self.op_manager = get_global_op_manager()

        # Register the op with the op_manager.
        # Only registers the op if op_id is not None.
        self.op_manager.store_op(op=self)

    def log_time_taken(self, checkpoint: float, _msg_i: int):

        now = time.time()
        memory_stats = psutil.virtual_memory()
        memory_used = memory_stats.used / 1024**2
        memory_used_percent = memory_stats.percent
        logger.info(
            f"Op.run: Time taken for "
            f"{self.op_obj.OP_FAMILY=} | {self.op_type=} | {self.op_id=} | "
            f"{_msg_i=} | {now - checkpoint:0.3f} seconds | "
            f"{memory_used=:,.2f} MB | {memory_used_percent=:.2f}%"
        )

    def run(
        self,
        time_step: Optional[TimeStep] = None,
        msg: Optional[OpMsg] = None,
    ) -> Generator[Union[OpMsg, OpTimeStepMsg], None, None]:
        """Run the operator."""
        logger.info("-" * 40)
        logger.info(
            f"Op.run: Running | "
            f"{self.op_obj.OP_FAMILY=} | {self.op_type=} | {self.op_id=}"
        )

        checkpoint = time.time()

        if self.op_obj.OP_FAMILY == "group":
            self.op_obj: BaseGroupOp
            for _msg_i, _msg in enumerate(
                self.op_obj.run(time_step=time_step, msg=msg)
            ):
                self.log_time_taken(checkpoint=checkpoint, _msg_i=_msg_i)
                if self.store_metadata:
                    self.op_manager.store_metadata(op=self, msg=_msg)
                if self.store_data:
                    self.op_manager.store_data(op=self, msg=_msg)
                yield _msg
                checkpoint = time.time()

        elif self.op_obj.OP_FAMILY == "time_step":
            self.op_obj: BaseTimeStepOp
            for _msg_i, _msg in enumerate(self.op_obj.run()):
                self.log_time_taken(checkpoint=checkpoint, _msg_i=_msg_i)
                if not isinstance(_msg.data, TimeStep):
                    raise ValueError(
                        f"Expected msg.data to be of class TimeStep for BaseScheduleOp "
                        f"instead received {type(_msg.data)=}."
                    )
                if self.store_metadata:
                    self.op_manager.store_metadata(op=self, msg=_msg)
                if self.store_data:
                    self.op_manager.store_data(op=self, msg=_msg)
                yield _msg
                checkpoint = time.time()

        elif self.op_obj.OP_FAMILY == "connector":
            self.op_obj: BaseConnectorOp
            _msg = self.op_obj.run()
            self.log_time_taken(checkpoint=checkpoint, _msg_i=0)
            # ALWAYS STORE DATA FOR CONNECTOR
            self.op_manager.store_data(op=self, msg=_msg)
            if self.store_metadata:
                self.op_manager.store_metadata(op=self, msg=_msg)
            yield _msg
            checkpoint = time.time()

        elif self.op_obj.OP_FAMILY == "generator":
            self.op_obj: BaseGeneratorOp
            for _msg_i, _msg in enumerate(
                self.op_obj.run(time_step=time_step, msg=msg)
            ):
                self.log_time_taken(checkpoint=checkpoint, _msg_i=_msg_i)
                # Register output metadata with the op_manager.
                if self.store_metadata:
                    self.op_manager.store_metadata(op=self, msg=_msg)
                if self.store_data:
                    self.op_manager.store_data(op=self, msg=_msg)
                yield _msg
                checkpoint = time.time()

        elif self.op_obj.OP_FAMILY == "consumer":
            self.op_obj: BaseConsumerOp
            _msg = self.op_obj.run(time_step=time_step, msg=msg)
            self.log_time_taken(checkpoint=checkpoint, _msg_i=0)
            if self.store_metadata:
                self.op_manager.store_metadata(op=self, msg=_msg)
            if self.store_data:
                self.op_manager.store_data(op=self, msg=_msg)

        else:
            raise NotImplementedError()



import hashlib
import json
import logging
import time
import uuid
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple, Union

import bcrypt
import pandas as pd
from pydantic import StrictBool, StrictInt, StrictStr, conlist, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class MailchimpGenericComputeEmailHashesOpConfigModel(BaseOpConfigModel):
    compute_hash_md5: StrictBool = True
    compute_hash_uuid5: StrictBool = True
    compute_hash_sha256: StrictBool = True
    compute_hash_sha512: StrictBool = True
    compute_hash_bcrypt: StrictBool = False
    bcrypt_salt_work_factor: StrictInt = 4

    @root_validator(pre=False)
    def validate_bcrypt_config(cls, values):
        # If compute_hash_bcrypt is True, bcrypt_salt_work_factor must be set.
        if values.get("compute_hash_bcrypt") is True and not isinstance(
            values.get("bcrypt_salt_work_factor", None), int
        ):
            raise ValueError(
                "If compute_hash_bcrypt is True, bcrypt_salt_work_factor must be set."
            )
        return values

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MailchimpGenericComputeEmailHashesOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MailchimpGenericComputeEmailHashesOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MailchimpGenericComputeEmailHashesOp(BaseGeneratorOp):
    """."""

    OP_TYPE = "generator.mailchimp.generic.compute_email_hashes"
    OP_CONFIG_MODEL = MailchimpGenericComputeEmailHashesOpConfigModel
    OP_METADATA_MODEL = MailchimpGenericComputeEmailHashesOpMetadataModel
    OP_AUDIT_MODEL = MailchimpGenericComputeEmailHashesOpAuditModel

    templated_fields = None

    def __init__(
        self,
        compute_hash_md5: bool = True,
        compute_hash_uuid5: bool = True,
        compute_hash_sha256: bool = True,
        compute_hash_sha512: bool = True,
        compute_hash_bcrypt: bool = False,
        bcrypt_salt_work_factor: int = 4,
        **kwargs,
    ):
        """."""
        self.compute_hash_md5 = compute_hash_md5
        self.compute_hash_uuid5 = compute_hash_uuid5
        self.compute_hash_sha256 = compute_hash_sha256
        self.compute_hash_sha512 = compute_hash_sha512
        self.compute_hash_bcrypt = compute_hash_bcrypt
        self.bcrypt_salt_work_factor = bcrypt_salt_work_factor
        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"MailchimpGenericComputeEmailHashesOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="MailchimpGenericComputeEmailHashesOp.run:",
        )

        input_records = msg.data

        if not isinstance(input_records, list):
            raise NotImplementedError("Input records must be a list of dicts.")

        final_records = []

        for record in input_records:
            if "email_address" not in record["data"]:
                raise NotImplementedError(
                    "Input records must have an email_address field."
                )

            data = record["data"]

            if self.compute_hash_md5:
                data["email_address_hash_md5"] = hashlib.md5(
                    data["email_address"].encode("utf-8")
                ).hexdigest()

            if self.compute_hash_uuid5:
                data["email_address_hash_uuid5"] = str(
                    uuid.uuid5(uuid.NAMESPACE_URL, data["email_address"])
                )

            if self.compute_hash_sha256:
                data["email_address_hash_sha256"] = hashlib.sha256(
                    data["email_address"].encode("utf-8")
                ).hexdigest()

            if self.compute_hash_sha512:
                data["email_address_hash_sha512"] = hashlib.sha512(
                    data["email_address"].encode("utf-8")
                ).hexdigest()

            if self.compute_hash_bcrypt:
                wf = self.bcrypt_salt_work_factor
                data[f"email_address_hash_bcrypt_wf_{wf}"] = bcrypt.hashpw(
                    data["email_address"].encode("utf-8"),
                    bcrypt.gensalt(rounds=self.bcrypt_salt_work_factor),
                ).decode("utf-8")

            # record['data'] has been updated in place.
            final_records.append(record)

        yield OpMsg(
            data=final_records,
            metadata=MailchimpGenericComputeEmailHashesOpMetadataModel(),
            audit=MailchimpGenericComputeEmailHashesOpAuditModel(),
        )



import hashlib
import json
import logging
import time
import uuid
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple, Union

import bcrypt
import pandas as pd
from pydantic import StrictBool, StrictInt, StrictStr, conlist, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class MailchimpGenericRemoveLinksOpConfigModel(BaseOpConfigModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MailchimpGenericRemoveLinksOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MailchimpGenericRemoveLinksOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MailchimpGenericRemoveLinksOp(BaseGeneratorOp):
    """."""

    OP_TYPE = "generator.mailchimp.generic.remove_links"
    OP_CONFIG_MODEL = MailchimpGenericRemoveLinksOpConfigModel
    OP_METADATA_MODEL = MailchimpGenericRemoveLinksOpMetadataModel
    OP_AUDIT_MODEL = MailchimpGenericRemoveLinksOpAuditModel

    templated_fields = None

    def __init__(
        self,
        **kwargs,
    ):
        """."""
        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"MailchimpGenericRemoveLinksOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="MailchimpGenericRemoveLinksOp.run:",
        )

        input_records = msg.data

        if not isinstance(input_records, list):
            raise NotImplementedError("Input records must be a list of dicts.")

        final_records = []

        logger.info(
            f"MailchimpGenericRemoveLinksOp.run: Redacting links from the data."
        )
        # Remove links from records
        # NOTE: Links can cause issues with deduplication.
        for record in input_records:
            data = record["data"]
            if "archive_url" in data:
                data["archive_url"] = "REDACTED_BY_ETL_BEFORE_STORAGE"
            if "long_archive_url" in data:
                data["long_archive_url"] = "REDACTED_BY_ETL_BEFORE_STORAGE"
            if "subscribe_url_short" in data:
                data["subscribe_url_short"] = "REDACTED_BY_ETL_BEFORE_STORAGE"
            if "subscribe_url_long" in data:
                data["subscribe_url_long"] = "REDACTED_BY_ETL_BEFORE_STORAGE"
            if "_links" in data:
                data["_links"] = "REDACTED_BY_ETL_BEFORE_STORAGE"

            # record['data'] has been updated in place.
            final_records.append(record)

        logger.info(
            f"MailchimpGenericRemoveLinksOp.run: Done redacting links from the data."
        )

        yield OpMsg(
            data=final_records,
            metadata=MailchimpGenericRemoveLinksOpMetadataModel(),
            audit=MailchimpGenericRemoveLinksOpAuditModel(),
        )



import json
import logging
import time
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple

import pandas as pd
from pydantic import StrictBool, StrictInt, StrictStr, conlist
from zeep import Client
from zeep.helpers import serialize_object

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp

logger = logging.getLogger(__name__)


class OngoingOrdersSOAPGetAllOrdersOpConfigModel(BaseOpConfigModel):
    http_conn_id: StrictStr
    goods_owner_code: StrictStr
    sleep_time_between_pages: int = 5
    min_date_modified: Optional[StrictStr] = None
    max_date_modified: Optional[StrictStr] = None
    iterate_over_pages: StrictBool
    orders_per_page: StrictInt = 1000

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OngoingOrdersSOAPGetAllOrdersOpMetadataModel(BaseOpMetadataModel):
    current_page: int
    order_id_from: int
    total_pages: int
    total_records: int

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OngoingOrdersSOAPGetAllOrdersOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OngoingOrdersSOAPGetAllOrdersOp(BaseGeneratorOp):
    """."""

    OP_TYPE = "generator.ongoing.orders.soap_get_all_orders"
    OP_CONFIG_MODEL = OngoingOrdersSOAPGetAllOrdersOpConfigModel
    OP_METADATA_MODEL = OngoingOrdersSOAPGetAllOrdersOpMetadataModel
    OP_AUDIT_MODEL = OngoingOrdersSOAPGetAllOrdersOpAuditModel

    templated_fields = None

    def __init__(
        self,
        http_conn_id: str,
        goods_owner_code: str,
        iterate_over_pages: bool,
        sleep_time_between_pages: int = 5,
        min_date_modified: Optional[StrictStr] = None,
        max_date_modified: Optional[StrictStr] = None,
        orders_per_page: int = 1000,
        **kwargs,
    ):
        """."""
        self.http_conn_id = http_conn_id
        self.goods_owner_code = goods_owner_code
        self.sleep_time_between_pages = sleep_time_between_pages
        self.min_date_modified = min_date_modified
        self.max_date_modified = max_date_modified
        self.iterate_over_pages = iterate_over_pages
        self.orders_per_page = orders_per_page

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.http_conn: HTTPConnOp = op_manager.get_connector(op_id=self.http_conn_id)

        if not self.http_conn.base_url.endswith("/"):
            wdsl_url = self.http_conn.base_url + "/service.asmx?WSDL"
        else:
            raise NotImplementedError()

        self.zeep_client = Client(wsdl=wdsl_url)

    def get_page(self, order_id_from: int) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """."""
        params = {}

        params["OrderIdFrom"] = order_id_from

        if self.min_date_modified:
            params["LastUpdatedFrom"] = self.min_date_modified
        if self.max_date_modified:
            params["LastUpdatedTo"] = self.max_date_modified

        # DOCS: https://developer.ongoingwarehouse.com/get-orders-by-query
        logger.info(
            f"OngoingOrdersSOAPGetAllOrdersOp.get_page: Fetching page with "
            f"{order_id_from=}"
        )

        etl_request_start_ts = pd.Timestamp.now(tz="UTC").isoformat()
        response = self.zeep_client.service.GetOrdersByQuery(
            GoodsOwnerCode=self.goods_owner_code,
            UserName=self.http_conn.username,
            Password=self.http_conn.password,
            query=params,
        )
        etl_response_end_ts = pd.Timestamp.now(tz="UTC").isoformat()
        response_dict = serialize_object(response, dict)

        etl_metadata_json = {
            "etl_request_start_ts": etl_request_start_ts,
            "etl_response_end_ts": etl_response_end_ts,
        }

        return response_dict, etl_metadata_json

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"OngoingOrdersSOAPGetAllOrdersOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="OngoingOrdersSOAPGetAllOrdersOp.run:",
        )

        final_records = []
        page, total_records, total_pages, order_id_from = 1, 0, None, 1
        while True:
            logger.info(
                f"OngoingOrdersSOAPGetAllOrdersOp.run: Fetching {page=} and "
                f"{order_id_from=}."
            )

            output_json, etl_metadata_json = self.get_page(order_id_from=order_id_from)
            if output_json is None:
                logger.info(
                    f"OngoingOrdersSOAPGetAllOrdersOp.run: Received null Response. "
                    f"No data exists for current {page=}. Stopping the loop to fetch "
                    f"additional pages."
                )
                total_pages = page - 1
                break

            n_records = len(output_json["Order"])
            total_records += n_records

            logger.info(
                f"OngoingOrdersSOAPGetAllOrdersOp.run: Done fetching {page=} and "
                f"{order_id_from=}. "
                f"Fetched {n_records=} this request. "
                f"Fetched {total_records=} records so far. "
            )

            ############################################################################
            # NOTE: Adding `etl_metadata` into response json (output_json).
            # * This is to add some metadata about the ETL process, primarily some
            #   timestamps which may help with creating event stores/logs.
            # * This may also help with incremental modelling.
            # * DO NOT USE this information when calculating dedup_uuid.
            ############################################################################
            etl_metadata_json["time_step"] = time_step.to_json_dict()
            records = [
                {
                    "data": record,
                    "etl_metadata": etl_metadata_json,
                }
                for record in output_json["Order"]
            ]
            final_records.extend(records)

            # Compute next order_id_from
            order_id_from = max(
                [record["data"]["OrderInfo"]["OrderId"] for record in records]
            )
            order_id_from += 1

            # Yield current page's results
            # If true, yield current records. Afterwards set final_records to empty list
            # If true, make sure to not yield records outside of loop.
            if self.iterate_over_pages:
                yield OpMsg(
                    data=final_records,
                    metadata=OngoingOrdersSOAPGetAllOrdersOpMetadataModel(
                        current_page=page,
                        total_pages=-1,
                        total_records=total_records,
                        order_id_from=order_id_from,
                    ),
                    audit=OngoingOrdersSOAPGetAllOrdersOpAuditModel(),
                )
                final_records = []

            logger.info(
                f"OngoingOrdersSOAPGetAllOrdersOp.run: Sleeping for "
                f"{self.sleep_time_between_pages} seconds between pages. "
                f"Next page is {page+1} and {order_id_from=}."
            )
            time.sleep(self.sleep_time_between_pages)

            # IMPORTANT: The page number is 1-indexed.
            page += 1

        # Only yield records if iterate_over_pages is false.
        if self.iterate_over_pages:
            logger.info(
                f"OngoingOrdersSOAPGetAllOrdersOp.run: {self.iterate_over_pages=}. "
                f"Finished yielding all records, once per page."
            )

        else:
            logger.info(
                f"OngoingOrdersSOAPGetAllOrdersOp.run: {self.iterate_over_pages=}. "
                f"Yielding all pages at once now. {len(final_records)=}"
            )
            yield OpMsg(
                data=final_records,
                metadata=OngoingOrdersSOAPGetAllOrdersOpMetadataModel(
                    current_page=page,
                    total_pages=total_pages,
                    total_records=total_records,
                    order_id_from=order_id_from,
                ),
                audit=OngoingOrdersSOAPGetAllOrdersOpAuditModel(),
            )



"""."""
import logging
import queue

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class PythonQueueHookConnOpConfigModel(BaseOpConfigModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonQueueHookConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonQueueHookConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonQueueHookConnOp(BaseConnectorOp):
    OP_TYPE = "connector.hooks.python.queue"
    OP_CONFIG_MODEL = PythonQueueHookConnOpConfigModel
    OP_METADATA_MODEL = PythonQueueHookConnOpMetadataModel
    OP_AUDIT_MODEL = PythonQueueHookConnOpAuditModel

    templated_fields = None

    def __init__(self, **kwargs):
        self.templated_fields = kwargs.get("templated_fields", [])

        # NOTE: This is the list :)
        self._queue = queue.Queue()

    def run(self) -> OpMsg:
        """."""
        logger.info(f"PythonQueueHookConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="PythonQueueHookConnOp.run:"
        )

        return OpMsg(
            data=self,
            metadata=PythonQueueHookConnOpMetadataModel(),
            audit=PythonQueueHookConnOpAuditModel(),
        )

    def get_reference(self):
        """."""
        return self._queue

    def get_reference_and_reinit(self):
        """."""
        _queue = self._queue
        self._queue = queue.Queue()
        return _queue

    def put(self, item):
        """."""
        self._queue.put(item, block=True)

    def get(self):
        """."""
        self._queue.get(block=True)



from typing import Any

from pydantic import BaseModel

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.time_step import TimeStep


class OpMsg(BaseModel):
    data: Any
    metadata: BaseOpMetadataModel
    audit: BaseOpAuditModel

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OpTimeStepMsg(OpMsg):
    data: TimeStep
